@@ -39,23 +39,35 @@ class InventoryBatchHandler(BatchHandler):
     """
     batch_model_class = model.InventoryBatch
 
+    def init_batch(self, batch, **kwargs):
+        if batch.total_cost is None:
+            batch.total_cost = 0
+
     def should_populate(self, batch):
-        # must populate only if we come from handheld batch(es)
-        return bool(batch.handheld_batches)
+        if batch.handheld_batches:
+            return True
+        if batch.mode == self.enum.INVENTORY_MODE_ZERO_ALL:
+            return True
 
     def populate(self, batch, progress=None):
         """
         Pre-fill batch with row data from an input data file, parsed according
         to the batch device type.
         """
-        if not batch.handheld_batches:
-            raise ValueError("Inventory batch does not have data source from which to populate")
+        if batch.handheld_batches:
+            self.populate_from_handheld(batch, progress=progress)
+        elif batch.mode == self.enum.INVENTORY_MODE_ZERO_ALL:
+            self.populate_zero_all(batch, progress=progress)
+
+    def populate_from_handheld(self, batch, progress=None):
 
         def append(hh, i):
             row = model.InventoryBatchRow()
             row.upc = hh.upc
-            row.cases = hh.cases or None
-            row.units = hh.units or None
+            if hh.cases is not None:
+                row.cases = hh.cases
+            if hh.units is not None:
+                row.units = hh.units
             self.add_row(batch, row)
 
         data = []
@@ -64,8 +76,32 @@ class InventoryBatchHandler(BatchHandler):
         self.progress_loop(append, data, progress,
                            message="Adding initial rows to batch")
 
+    def populate_zero_all(self, batch, progress=None):
+        session = orm.object_session(batch)
+        products = session.query(model.Product)\
+                          .join(model.ProductInventory)\
+                          .filter(model.ProductInventory.on_hand != None)\
+                          .filter(model.ProductInventory.on_hand != 0)
+
+        def append(product, i):
+            row = model.InventoryBatchRow()
+            row.product = product
+            row.units = 0
+            self.add_row(batch, row)
+
+        self.progress_loop(append, products, progress,
+                           message="Adding initial rows to batch")
+
     def refresh(self, batch, progress=None):
         batch.total_cost = 0
+
+        # destroy and re-create data rows if batch is zero-all
+        if batch.mode == self.enum.INVENTORY_MODE_ZERO_ALL:
+            del batch.data_rows[:]
+            batch.rowcount = 0
+            self.populate_zero_all(batch, progress=progress)
+            return True
+
         return super(InventoryBatchHandler, self).refresh(batch, progress=progress)
 
     def refresh_row(self, row):
@@ -92,14 +128,33 @@ class InventoryBatchHandler(BatchHandler):
         row.status_code = row.STATUS_OK
         row.case_quantity = (product.cost.case_size or 1) if product.cost else 1
 
-        # dynamic?
-        row.unit_cost = self.get_unit_cost(row)
+        if row.previous_units_on_hand is not None:
+            row.variance = self.total_units(row) - row.previous_units_on_hand
+
+        # TODO: is this a sufficient check?  need to avoid overwriting a cost
+        # value which has been manually set, but this also means the first
+        # value that lands will stick, and e.g. new cost would be ignored
+        if row.unit_cost is None:
+            row.unit_cost = self.get_unit_cost(row)
+
         self.refresh_totals(row)
 
+    def total_units(self, row):
+        return (row.cases or 0) * (row.case_quantity or 1) + (row.units or 0)
+
+    def capture_current_units(self, row):
+        """
+        Capture the "current" (aka. "previous") unit count for the row, if
+        applicable.
+        """
+        product = row.product
+        if product and product.inventory:
+            row.previous_units_on_hand = product.inventory.on_hand
+
     def refresh_totals(self, row):
         batch = row.batch
 
-        if row.unit_cost:
+        if row.unit_cost is not None:
             row.total_cost = row.unit_cost * (row.full_unit_quantity or 0)
             batch.total_cost = (batch.total_cost or 0) + row.total_cost
         else:
