@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2019 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -27,6 +27,7 @@ DataSync Consumers
 from __future__ import unicode_literals, absolute_import
 
 from rattail import importing
+from rattail.db import Session
 from rattail.config import parse_list
 from rattail.util import load_object
 
@@ -35,12 +36,16 @@ class DataSyncConsumer(object):
     """
     Base class for all DataSync consumers.
     """
+    delay = 1 # seconds
+    retry_attempts = 1
+    retry_delay = 1 # seconds
 
-    def __init__(self, config, key, dbkey=None, runas=None):
+    def __init__(self, config, key, dbkey=None, runas=None, watcher=None):
         self.config = config
         self.key = key
         self.dbkey = dbkey
         self.runas_username = runas
+        self.watcher = watcher
 
     def setup(self):
         """
@@ -80,10 +85,19 @@ class NewDataSyncImportConsumer(DataSyncConsumer):
        ``rattail.importing.Importer``.
 
     .. attribute:: handler_spec
+
        This should be a "spec" string referencing the import handler class from
        which the importers should be obtained.
 
+    .. attribute:: default_importers_only
+
+       Flag indicating whether the consumer should *only* use "default"
+       importers from the handler, vs. using "all" importers, when processing
+       changes.  This flag is ``True`` by default, meaning only "default"
+       importers will be used.
+
     .. attribute:: model_map
+
        This is a dictionary which may be used to map model names (importer
        "keys") between the host and local systems.  Keys of the dictionary
        should be the model name as it comes from the host system, in the form
@@ -92,9 +106,11 @@ class NewDataSyncImportConsumer(DataSyncConsumer):
        handler, i.e. in its ``importer_keys()`` return value.
 
     .. attribute:: skip_local_models
+
        This is a list of model names, as they are found in the import handler.
     """
     handler_spec = None
+    default_importers_only = True
     model_map = {}
     skip_local_models = []
 
@@ -105,24 +121,30 @@ class NewDataSyncImportConsumer(DataSyncConsumer):
 
     def get_importers(self):
         """
-        Returns a dictionary, keys of which are *generally* model names
+        Returns a dictionary, keys of which are "model" names
         (e.g. ``'Product'``) and values of which are
-        :class:`rattail.importing.Importer` instances.
+        :class:`~rattail.importing.importers.Importer` instances.
 
         .. note::
-           The keys must ultimately align with the ``change.payload_type``
-           values, coming from the host system.  Or at least that's what
-           will make life easy for the :meth:`invoke_importer()` method.
+           The keys must ultimately align with the
+           :attr:`~rattail.db.model.datasync.DataSyncChange.payload_type`
+           values, coming from the host system.  Or at least that's what will
+           make life easy for the :meth:`invoke_importer()` method.
         """
         handler = load_object(self.handler_spec)
-        importers = self.get_importers_from_handler(handler, default_only=True)
-        for host_name, local_name in self.model_map.iteritems():
+
+        importers = self.get_importers_from_handler(
+            handler, default_only=self.default_importers_only)
+
+        for host_name, local_name in self.model_map.items():
             if local_name in importers:
                 importers[host_name] = importers[local_name]
+
         if self.skip_local_models:
             for name in list(importers):
                 if name in self.skip_local_models:
                     del importers[name]
+
         return importers
 
     def get_importers_from_handler(self, handler, default_only=True):
@@ -134,19 +156,19 @@ class NewDataSyncImportConsumer(DataSyncConsumer):
         else:
             keys = handler.get_importer_keys()
 
-        importers = {}
-        factories = handler.get_importers()
-        for key in keys:
-            importers[key] = factories[key](config=self.config)
-
+        importers = dict([(key, handler.get_importer(key))
+                          for key in keys])
         return importers
 
     def process_changes(self, session, changes):
         """
         Process all changes, leveraging importer(s) as much as possible.
         """
+        if self.runas_username:
+            session.set_continuum_user(self.runas_username)
+
         # Update all importers with current Rattail session.
-        for importer in self.importers.itervalues():
+        for importer in self.importers.values():
             importer.session = session
 
         for change in changes:
@@ -177,7 +199,7 @@ class NewDataSyncImportConsumer(DataSyncConsumer):
             if host_data is None:
                 return
         key = importer.get_key(host_data)
-        local_object = importer.get_local_object(key)
+        local_object = self.get_local_object(importer, key, host_data)
         if local_object:
             if importer.allow_update:
                 local_data = importer.normalize_local_object(local_object)
@@ -211,6 +233,22 @@ class NewDataSyncImportConsumer(DataSyncConsumer):
         """
         raise NotImplementedError
 
+    def get_local_object(self, importer, key, host_data):
+        """
+        Fetch the "local" (destination) object for the given key.
+        """
+        return importer.get_local_object(key)
+
+
+class NullTestConsumer(DataSyncConsumer):
+    """
+    Consumer which simply does ignores changes, letting DataSync think they
+    were processed okay.
+    """
+
+    def process_changes(self, session, changes):
+        pass
+
 
 class ErrorTestConsumer(DataSyncConsumer):
     """
