@@ -26,45 +26,44 @@ Authentication & Authorization
 
 from __future__ import unicode_literals, absolute_import
 
-import bcrypt
-import six
+from passlib.context import CryptContext
+from sqlalchemy.orm.exc import NoResultFound
 
 from rattail.db import model
 
 
-def authenticate_user(session, username, password):
+password_context = CryptContext(schemes=['bcrypt'])
+
+
+def authenticate_user(session, userobj, password):
     """
     Attempt to authenticate a user.
 
-    :returns: A :class:`.model.User` instance, if successful.  Otherwise
-       returns ``None``.
+    :param userobj: May be a :class:`model.User` instance, or a username as
+       string.  If the latter, it will be used to look up the User instance.
+
+    :returns: The User instance, if found and the password was correct;
+       otherwise ``None``.
     """
-    user = session.query(model.User)\
-        .filter_by(username=username)\
-        .first()
-    if user and user.active and user.password is not None:
-        # apparently bcrypt's hashpw() doesn't like Unicode
-        if isinstance(password, six.text_type):
-            password = password.encode('utf_8')
-        salt = user.salt.encode('utf_8')
+    if isinstance(userobj, model.User):
+        user = userobj
+    else:
         try:
-            authenticated = (bcrypt.hashpw(password, salt) == user.password)
-        except UnicodeEncodeError:
-            authenticated = False
-        if authenticated:
+            user = session.query(model.User)\
+                          .filter_by(username=userobj)\
+                          .one()
+        except NoResultFound:
+            user = None
+    if user and user.active and user.password is not None:
+        if password_context.verify(password, user.password):
             return user
-    return None
 
 
 def set_user_password(user, password):
     """
     Set a user's password.
     """
-    # apparently bcrypt's hashpw() doesn't like Unicode
-    if isinstance(password, six.text_type):
-        password = password.encode('utf_8')
-    user.salt = bcrypt.gensalt()
-    user.password = bcrypt.hashpw(password, user.salt)
+    user.password = password_context.hash(password)
 
 
 def special_role(session, uuid, name):
@@ -155,3 +154,55 @@ def has_permission(session, principal, permission, include_guest=True, include_a
             if perm == permission:
                 return True
     return False
+
+
+def cache_permissions(session, principal, include_guest=True, include_authenticated=True):
+    """
+    Return a set of permission names, which represents all permissions
+    effectively granted to the given principal.
+
+    :param session: A SQLAlchemy session instance.
+
+    :param principal: May be either a :class:`.model.User` or
+       :class:`.model.Role` instance.  It is also expected that this may
+       sometimes be ``None``, in which case the "Guest" role will typically be
+       assumed.
+
+    :param include_guest: Whether or not the "Guest" role should be included
+       when checking permissions.  If ``False``, then Guest's permissions will
+       *not* be consulted.
+
+    :param include_authenticated: Whether or not the "Authenticated" role
+       should be included when checking permissions.
+
+    Note that if no ``principal`` is provided, and ``include_guest`` is set to
+    ``False``, then no checks will actually be done, and the return value will
+    be ``False``.
+    """
+    # we will use any `roles` attribute which may be present.  in practice we
+    # would be assuming a User in this case
+    if hasattr(principal, 'roles'):
+        roles = list(principal.roles)
+
+        # here our User assumption gets a little more explicit
+        if include_authenticated:
+            roles.append(authenticated_role(session))
+
+    # otherwise a non-null principal is assumed to be a Role
+    elif principal is not None:
+        roles = [principal]
+
+    # fallback assumption is "no roles"
+    else:
+        roles = []
+
+    # maybe include guest roles
+    if include_guest:
+        roles.append(guest_role(session))
+
+    # build the permissions cache
+    cache = set()
+    for role in roles:
+        cache.update(role.permissions)
+
+    return cache
