@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2019 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -26,23 +26,26 @@ Rattail data models
 
 from __future__ import unicode_literals, absolute_import
 
-from .core import Base, ModelBase, uuid_column, getset_factory, GPCType, Setting, Change
+from .core import Base, ModelBase, uuid_column, getset_factory, GPCType, Setting, Change, Note
 from .contact import PhoneNumber, EmailAddress, MailingAddress
 
-from .people import Person, PersonPhoneNumber, PersonEmailAddress, PersonMailingAddress
+from .people import Person, PersonPhoneNumber, PersonEmailAddress, PersonMailingAddress, PersonNote
 from .users import Role, Permission, User, UserRole, UserEvent
 from .stores import Store, StorePhoneNumber, StoreEmailAddress
 from .customers import (Customer, CustomerPhoneNumber, CustomerEmailAddress, CustomerMailingAddress,
-                        CustomerGroup, CustomerGroupAssignment, CustomerPerson)
+                        CustomerGroup, CustomerGroupAssignment, CustomerPerson, CustomerNote)
+from .members import Member, MemberPhoneNumber, MemberEmailAddress, MemberMailingAddress
 
 from .org import Department, Subdepartment, Category, Family, ReportCode, DepositLink
 from .employees import (Employee, EmployeePhoneNumber, EmployeeEmailAddress,
-                        EmployeeStore, EmployeeDepartment)
+                        EmployeeStore, EmployeeDepartment, EmployeeHistory)
 from .shifts import ScheduledShift, WorkedShift
 
 from .vendors import Vendor, VendorPhoneNumber, VendorEmailAddress, VendorContact
-from .products import (Brand, Tax, Product, ProductImage, ProductCode, ProductCost, ProductPrice,
-                       ProductInventory, ProductStoreInfo)
+from .products import (Brand, Tax, Product, ProductImage, ProductCode,
+                       ProductCost, ProductFutureCost, ProductPrice,
+                       ProductInventory, ProductStoreInfo, ProductVolatile,
+                       InventoryAdjustmentReason)
 from .purchase import (PurchaseBase, PurchaseItemBase, PurchaseCreditBase,
                        Purchase, PurchaseItem, PurchaseCredit)
 
@@ -52,16 +55,20 @@ from .messages import Message, MessageRecipient
 
 from .datasync import DataSyncChange
 from .labels import LabelProfile
-from .bouncer import EmailBounce
+from .bouncer import EmailAttempt, EmailBounce
 from .tempmon import TempmonClient, TempmonProbe, TempmonReading
+from .upgrades import Upgrade, UpgradeRequirement
 
 from .exports import ExportMixin
 from .reports import ReportOutput
 from .batch import BatchMixin, BaseFileBatchMixin, FileBatchMixin, BatchRowMixin, ProductBatchRowMixin
+from .batch.dynamic import DynamicBatchMixin, ImporterBatch
 from .batch.handheld import HandheldBatch, HandheldBatchRow
 from .batch.inventory import InventoryBatch, InventoryBatchRow
 from .batch.labels import LabelBatch, LabelBatchRow
+from .batch.newproduct import NewProductBatch, NewProductBatchRow
 from .batch.pricing import PricingBatch, PricingBatchRow
-from .batch.purchase import PurchaseBatch, PurchaseBatchRow, PurchaseBatchCredit
+from .batch.product import ProductBatch, ProductBatchRow
+from .batch.purchase import PurchaseBatch, PurchaseBatchRow, PurchaseBatchRowClaim, PurchaseBatchCredit
 from .batch.vendorcatalog import VendorCatalog, VendorCatalogRow
 from .batch.vendorinvoice import VendorInvoice, VendorInvoiceRow
