@@ -1,8 +1,8 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; -*-
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2018 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -40,19 +40,10 @@ from sqlalchemy.orm import relationship, object_session
 from sqlalchemy.ext.declarative import declared_attr
 from sqlalchemy.ext.orderinglist import ordering_list
 
-from rattail.db.core import uuid_column
-from rattail.db.types import GPCType
+from rattail.db.core import uuid_column, filename_column
+from rattail.db.types import GPCType, JSONTextDict
 from rattail.db.model import User, Product
-
-
-def filename_column(**kwargs):
-    """
-    Return a SQLAlchemy Column object suitable for defining a model field.  It
-    is hoped this will be useful for standardization and/or readability..?
-    """
-    kwargs.setdefault('nullable', True)
-    kwargs.setdefault('doc', "Base name of the data file.")
-    return sa.Column(sa.String(length=255), **kwargs)
+from rattail.time import make_utc
 
 
 @six.python_2_unicode_compatible
@@ -91,6 +82,10 @@ class BatchMixin(object):
     Rattail database.
     """)
 
+    description = sa.Column(sa.String(length=255), nullable=True, doc="""
+    Basic (loosely identifying) description for the batch.
+    """)
+
     created = sa.Column(sa.DateTime(), nullable=False, default=datetime.datetime.utcnow, doc="""
     Date and time when the batch was first created.
     """)
@@ -127,7 +122,7 @@ class BatchMixin(object):
     Cached row count for the batch.  No guarantees perhaps, but should be accurate.
     """)
 
-    complete = sa.Column(sa.Boolean(), nullable=True, doc="""
+    complete = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
     Flag to indicate whether the batch is complete.  This may be used to assist
     with workflow when entering/executing new batches.
     """)
@@ -156,6 +151,17 @@ class BatchMixin(object):
     Any arbitrary notes for the batch.
     """)
 
+    params = sa.Column(JSONTextDict(), nullable=True, doc="""
+    Extra parameters for the batch, encoded as JSON.  This hopefully can be
+    useful for "many" batches, to avoid needing to add extra flags to the
+    schema etc.
+    """)
+
+    extra_data = sa.Column(sa.Text(), nullable=True, doc="""
+    Extra field for arbitrary data, useful to the batch handler etc.  Note that
+    there is *no* structure assumed here, it can be JSON or whatever is needed.
+    """)
+
     STATUS_OK                   = 1
     STATUS_QUESTIONABLE         = 2
 
@@ -362,6 +368,11 @@ class BatchRowMixin(object):
     defined by each particular batch type.
     """)
 
+    modified = sa.Column(sa.DateTime(), nullable=True, default=make_utc, onupdate=make_utc, doc="""
+    Last modification time of the row.  This should be automatically set when
+    the row is first created, as well as anytime it's updated thereafter.
+    """)
+
     removed = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
     Flag to indicate a row has been removed from the batch.
     """)
@@ -383,6 +394,13 @@ class ProductBatchRowMixin(BatchRowMixin):
                                     name='{0}_fk_product'.format(row_table)),
             )
 
+    item_entry = sa.Column(sa.String(length=32), nullable=True, doc="""
+    Raw entry value, as obtained from the initial data source, and which is
+    used to locate the product within the system.  This raw value is preserved
+    in case the initial lookup fails and a refresh must attempt further
+    lookup(s) later.  Only used by certain batch handlers in practice.
+    """)
+
     upc = sa.Column(GPCType(), nullable=True, doc="""
     UPC for the product associated with the row.
     """)
@@ -432,3 +450,11 @@ class ProductBatchRowMixin(BatchRowMixin):
     department_name = sa.Column(sa.String(length=30), nullable=True, doc="""
     Name of the department to which the product belongs.
     """)
+
+    subdepartment_number = sa.Column(sa.Integer(), nullable=True, doc="""
+    Number of the subdepartment to which the product belongs.
+    """)
+
+    subdepartment_name = sa.Column(sa.String(length=30), nullable=True, doc="""
+    Name of the subdepartment to which the product belongs.
+    """)
