@@ -1,4 +1,4 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; -*-
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
@@ -35,6 +35,7 @@ from sqlalchemy.ext.declarative import declarative_base
 from sqlalchemy.ext.associationproxy import association_proxy
 
 from rattail.core import Object
+from rattail.time import make_utc
 
 # These are imported because most sibling modules import them *from* here (if
 # that makes sense).  Probably need to think harder about the "best" place for
@@ -50,7 +51,20 @@ class ModelBase(Object):
 
     def __repr__(self):
         mapper = orm.object_mapper(self)
-        keys = ['{0}={1}'.format(k.key, repr(getattr(self, k.key))) for k in mapper.primary_key]
+
+        key_props = []
+        for prop in mapper.column_attrs:
+            primary = True
+            for column in prop.columns:
+                if column not in mapper.primary_key:
+                    primary = False
+                    break
+            if primary:
+                key_props.append(prop)
+
+        keys = ['{}={}'.format(prop.key, repr(getattr(self, prop.key)))
+                for prop in key_props]
+
         return "{0}({1})".format(self.__class__.__name__, ', '.join(keys))
 
     @classmethod
@@ -127,3 +141,34 @@ class Change(Base):
     def __repr__(self):
         return "Change(class_name={}, instance_uuid={}, deleted={})".format(
             repr(self.class_name), repr(self.instance_uuid), repr(self.deleted))
+
+
+class Note(Base):
+    """
+    Arbitrary note attached to a supported object.
+    """
+    __tablename__ = 'note'
+    __table_args__ = (
+        sa.ForeignKeyConstraint(['created_by_uuid'], ['user.uuid'], name='note_fk_created_by'),
+    )
+    __versioned__= {}
+
+    uuid = uuid_column()
+    parent_type = sa.Column(sa.String(length=20), nullable=False)
+    parent_uuid = sa.Column(sa.String(length=32), nullable=False)
+    type = sa.Column(sa.String(length=15))
+    subject = sa.Column(sa.String(length=255), nullable=True)
+    text = sa.Column(sa.Text(), nullable=False)
+    created = sa.Column(sa.DateTime(), nullable=False, default=make_utc)
+
+    created_by_uuid = sa.Column(sa.String(length=32), nullable=False)
+    created_by = orm.relationship('User')
+
+    __mapper_args__ = {'polymorphic_on': parent_type}
+
+    def __str__(self):
+        return str(self.text)
+
+    if six.PY2:
+        def __unicode__(self):
+            return unicode(self.text)
