@@ -1,8 +1,8 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; -*-
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2020 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -31,10 +31,11 @@ import datetime
 import six
 import sqlalchemy as sa
 from sqlalchemy import orm
+from sqlalchemy.ext.declarative import declared_attr
 from sqlalchemy.ext.orderinglist import ordering_list
 
-from .core import Base, uuid_column
-from .contact import PhoneNumber, EmailAddress, MailingAddress
+from .core import Base, uuid_column, Note
+from .contact import PhoneNumber, EmailAddress, MailingAddress, ContactMixin
 from rattail.db.util import normalize_full_name
 
 
@@ -49,8 +50,7 @@ def get_person_display_name_from_context(context):
     return normalize_full_name(first_name, last_name)
 
 
-@six.python_2_unicode_compatible
-class Person(Base):
+class Person(ContactMixin, Base):
     """
     Represents a real, living and breathing person.
 
@@ -68,30 +68,88 @@ class Person(Base):
     last_name = sa.Column(sa.String(length=50), nullable=True)
     # TODO: rename this to 'full_name' at some point...?
     display_name = sa.Column(sa.String(length=100), nullable=True)
+
+    local_only = sa.Column(sa.Boolean(), nullable=True, doc="""
+    Flag indicating the person is somehow specific to the "local" app node
+    etc. and should not be synced elsewhere.
+    """)
+
+    invalid_address = sa.Column(sa.Boolean(), nullable=True, doc="""
+    Flag indicating the person's mailing address(es) on file are invalid.
+    """)
+
     modified = sa.Column(sa.DateTime(), nullable=True, onupdate=datetime.datetime.utcnow)
 
-    def __str__(self):
-        return self.display_name or ''
+    users = orm.relationship(
+        'User', doc="""
+        List of user accounts associated with this person.
+        """)
 
-    def add_email_address(self, address, type='Home'):
-        email = PersonEmailAddress(address=address, type=type)
-        self.emails.append(email)
+    def __str__(self):
+        if six.PY2:
+            return unicode(self).encode('utf8')
+        return self.display_name or ""
+
+    if six.PY2:
+        def __unicode__(self):
+            return self.display_name or ""
+
+    def add_email_address(self, address, type='Home', flush=False):
+        email = self.add_email(type=type, address=address, flush=flush)
         return email
 
-    def add_phone_number(self, number, type='Home'):
-        phone = PersonPhoneNumber(number=number, type=type)
-        self.phones.append(phone)
+    def add_phone_number(self, number, type='Home', flush=False):
+        phone = self.add_phone(type=type, number=number, flush=flush)
         return phone
 
+    def first_valid_email(self):
+        """
+        Returns the first :class:`Email` which has not been marked invalid, or ``None``.
+        """
+        for email in self.emails:
+            if not email.invalid:
+                return email
 
-class PersonPhoneNumber(PhoneNumber):
+    def first_valid_email_address(self):
+        """
+        Returns the first email address which has not been marked invalid, or ``None``.
+        """
+        email = self.first_valid_email()
+        if email:
+            return email.address
+
+    def only_customer(self):
+        """
+        Convenience method to retrieve the one and only ``Customer`` record
+        which is associated with this person.  An error will be raised if there
+        is not exactly one customer associated.
+        """
+        if len(self.customers) != 1:
+            raise ValueError("person should have 1 customer but instead has {}: {}".format(
+                len(self.customers), self))
+        return self.customers[0]
+
+
+class PersonContactInfoMixin(object):
+    """
+    Base mixin class for person contact info models.
+    """
+    Parent = Person
+
+    @declared_attr
+    def parent(cls):
+        return orm.synonym('person')
+
+
+class PersonPhoneNumber(PersonContactInfoMixin, PhoneNumber):
     """
     Represents a phone (or fax) number associated with a person.
     """
-
     __mapper_args__ = {'polymorphic_identity': 'Person'}
 
 
+Person._contact_phone_model = PersonPhoneNumber
+
 Person.phones = orm.relationship(
     PersonPhoneNumber,
     primaryjoin=PersonPhoneNumber.parent_uuid == Person.uuid,
@@ -122,14 +180,15 @@ Person.phone = orm.relationship(
     viewonly=True)
 
 
-class PersonEmailAddress(EmailAddress):
+class PersonEmailAddress(PersonContactInfoMixin, EmailAddress):
     """
     Represents an email address associated with a person.
     """
-
     __mapper_args__ = {'polymorphic_identity': 'Person'}
 
 
+Person._contact_email_model = PersonEmailAddress
+
 Person.emails = orm.relationship(
     PersonEmailAddress,
     primaryjoin=PersonEmailAddress.parent_uuid == Person.uuid,
@@ -160,14 +219,15 @@ Person.email = orm.relationship(
     viewonly=True)
 
 
-class PersonMailingAddress(MailingAddress):
+class PersonMailingAddress(PersonContactInfoMixin, MailingAddress):
     """
     Represents a physical / mailing address associated with a person.
     """
-
     __mapper_args__ = {'polymorphic_identity': 'Person'}
 
 
+Person._contact_address_model = PersonMailingAddress
+
 Person.addresses = orm.relationship(
     PersonMailingAddress,
     backref='person',
@@ -186,3 +246,26 @@ Person.address = orm.relationship(
     foreign_keys=[PersonMailingAddress.parent_uuid],
     uselist=False,
     viewonly=True)
+
+
+class PersonNote(Note):
+    """
+    Represents a note attached to a person.
+    """
+    __mapper_args__ = {'polymorphic_identity': 'Person'}
+
+
+Person.notes = orm.relationship(
+    PersonNote,
+    primaryjoin=PersonNote.parent_uuid == Person.uuid,
+    foreign_keys=[PersonNote.parent_uuid],
+    order_by=PersonNote.created,
+    cascade='all, delete-orphan',
+    doc="""
+    Sequence of notes which belong to the person.
+    """,
+    backref=orm.backref(
+        'person',
+        doc="""
+        Reference to the person to which the note is attached.
+        """))
