@@ -49,10 +49,23 @@ class Brand(Base):
     Represents a brand or similar product line.
     """
     __tablename__ = 'brand'
+    __table_args__ = (
+        sa.UniqueConstraint('name', name='brand_uq_name'),
+    )
     __versioned__ = {}
 
     uuid = uuid_column()
-    name = sa.Column(sa.String(length=100), nullable=True)
+
+    name = sa.Column(sa.String(length=100), nullable=False, doc="""
+    Brand name, as seen on product packaging etc.
+    """)
+
+    confirmed = sa.Column(sa.Boolean(), nullable=True, doc="""
+    Flag indicating the brand has been "confirmed good" by someone responsible
+    for product maintenance (if applicable).  In some cases the source of brand
+    data is a bit "loose" and it can be helpful to have a human review and
+    de-duplicate the brand listing.
+    """)
 
     def __str__(self):
         return self.name or ''
@@ -104,8 +117,11 @@ class Product(Base):
         sa.ForeignKeyConstraint(['report_code_uuid'], ['report_code.uuid'], name='product_fk_report_code'),
         sa.ForeignKeyConstraint(['deposit_link_uuid'], ['deposit_link.uuid'], name='product_fk_deposit_link'),
         sa.ForeignKeyConstraint(['tax_uuid'], ['tax.uuid'], name='product_fk_tax'),
+        sa.ForeignKeyConstraint(['suggested_price_uuid'], ['product_price.uuid'], name='product_fk_suggested_price', use_alter=True),
         sa.ForeignKeyConstraint(['regular_price_uuid'], ['product_price.uuid'], name='product_fk_regular_price', use_alter=True),
         sa.ForeignKeyConstraint(['current_price_uuid'], ['product_price.uuid'], name='product_fk_current_price', use_alter=True),
+        sa.ForeignKeyConstraint(['tpr_price_uuid'], ['product_price.uuid'], name='product_fk_tpr_price', use_alter=True),
+        sa.ForeignKeyConstraint(['sale_price_uuid'], ['product_price.uuid'], name='product_fk_sale_price', use_alter=True),
         sa.Index('product_ix_upc', 'upc'),
     )
     __versioned__ = {'exclude': ['last_sold']}
@@ -182,7 +198,7 @@ class Product(Base):
     the keys of the ``rattail.enum.UNIT_OF_MEASURE`` dictionary.
     """)
 
-    uom_abbreviation = sa.Column(sa.String(length=4), nullable=True, doc="""
+    uom_abbreviation = sa.Column(sa.String(length=10), nullable=True, doc="""
     Optional abbreviated unit of measure.  This value is intended to contain a
     human-friendly abbreviation for the :attr:`unit_of_measure`.  This value is
     generally also seen at the end of :attr:`size`.
@@ -210,6 +226,12 @@ class Product(Base):
     Number of *units* which constitute the current *pack* product, if applicable.
     """)
 
+    default_pack = sa.Column(sa.Boolean(), nullable=True, doc="""
+    If set, this flag indicates the product is the "default pack" for its unit
+    item.  This flag is only relevant if the product is in fact a pack item; it
+    should be effectively ignored for a unit item.
+    """)
+
     case_size = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
     Case size for the product, i.e. how many units (current product) are
     included when *sold* as a case.
@@ -253,8 +275,27 @@ class Product(Base):
     you use this is up to you.
     """)
 
-    regular_price_uuid = sa.Column(sa.String(length=32))
-    current_price_uuid = sa.Column(sa.String(length=32))
+    price_required = sa.Column(sa.Boolean(), nullable=True, doc="""
+    Indicates that the price must be manually entered by cashier, at check-out.
+    """)
+
+    suggested_price_uuid = sa.Column(sa.String(length=32), nullable=True)
+    regular_price_uuid = sa.Column(sa.String(length=32), nullable=True)
+    current_price_uuid = sa.Column(sa.String(length=32), nullable=True)
+
+    tpr_price_uuid = sa.Column(sa.String(length=32), nullable=True)
+    tpr_price = orm.relationship(
+        'ProductPrice',
+        primaryjoin='Product.tpr_price_uuid == ProductPrice.uuid',
+        lazy='joined',
+        post_update=True)
+
+    sale_price_uuid = sa.Column(sa.String(length=32), nullable=True)
+    sale_price = orm.relationship(
+        'ProductPrice',
+        primaryjoin='Product.sale_price_uuid == ProductPrice.uuid',
+        lazy='joined',
+        post_update=True)
 
     discountable = sa.Column(sa.Boolean(), nullable=False, default=True, doc="""
     Whether or not the product may be discounted in any way.
@@ -308,7 +349,16 @@ class Product(Base):
         Category,
         backref=orm.backref('products'))
 
-    brand = orm.relationship(Brand)
+    brand = orm.relationship(
+        Brand,
+        doc="""
+        Reference to the brand object, if any.
+        """,
+        backref=orm.backref(
+            'products',
+            doc="""
+            List of products for the brand.
+            """))
 
     family = orm.relationship(
         'Family',
@@ -353,6 +403,25 @@ class Product(Base):
         return make_full_description(self.brand.name if self.brand else '',
                                      self.description, self.size)
 
+    def is_unit_item(self):
+        """
+        Returns True if the product is considered a "unit" item, as opposed to
+        a "pack" item.
+        """
+        if self.unit:
+            # we appear to be a pack for some other unit item
+            return False
+        return True
+
+    def is_pack_item(self):
+        """
+        Returns True if the product is considered a "pack" item, as opposed to
+        a "unit" item.
+        """
+        if self.unit:
+            return True
+        return False
+
     @property
     def pretty_upc(self):
         """
@@ -362,14 +431,32 @@ class Product(Base):
             return None
         return self.upc.pretty()
 
-    def cost_for_vendor(self, vendor):
+    def costs_for_vendor(self, vendor):
         """
-        Of the product's cost records, return the first one associated with the
-        given vendor.
+        Return all costs associated with the given vendor.
         """
-        for cost in self.costs:
-            if cost.vendor is vendor:
-                return cost
+        return [cost for cost in self.costs
+                if cost.vendor is vendor]
+
+    def cost_for_vendor(self, vendor, error_if_multiple=False, error_if_none=False):
+        """
+        Locate and return the first cost associated with the given vendor.
+
+        :param error_if_multiple: If set, and the product has more than one
+           cost associated with the given vendor, an exception will be raised.
+
+        :param error_if_none: If set, and the product has no cost(s) associated
+           with the given vendor, an exception will be raised.
+        """
+        costs = self.costs_for_vendor(vendor)
+        if costs:
+            if error_if_multiple and len(costs) > 1:
+                raise ValueError("Product {} {} has {} costs associated with vendor: {}".format(
+                    self.upc, self.full_description, len(costs), vendor))
+            return costs[0]
+        elif error_if_none:
+            raise ValueError("Product {} {} has 0 costs associated with vendor: {}".format(
+                self.upc, self.full_description, vendor))
 
 
 class ProductImage(Base):
@@ -459,18 +546,57 @@ class ProductCost(Base):
     product_uuid = sa.Column(sa.String(length=32), nullable=False)
     vendor_uuid = sa.Column(sa.String(length=32), nullable=False)
     preference = sa.Column(sa.Integer(), nullable=False)
-    code = sa.Column(sa.String(length=20))
+
+    code = sa.Column(sa.String(length=20), nullable=True, doc="""
+    Vendor order code for the item, if applicable.
+    """)
 
     case_size = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
     Number of units which constitute a "case" in the context of this vendor
     catalog item.  May be a fractional quantity, e.g. 17.5LB.
     """)
 
-    case_cost = sa.Column(sa.Numeric(precision=9, scale=5))
+    # TODO: we're using precision=10, scale=5 elsewhere; should grow this too
+    case_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
+    Case cost for the item, if known.
+    """)
+
+    # TODO: we're using precision=10, scale=5 elsewhere; should grow this too
+    unit_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
+    Unit cost for the item, if known.
+    """)
+
     pack_size = sa.Column(sa.Integer())
     pack_cost = sa.Column(sa.Numeric(precision=9, scale=5))
-    unit_cost = sa.Column(sa.Numeric(precision=9, scale=5))
-    effective = sa.Column(sa.DateTime())
+
+    effective = sa.Column(sa.DateTime(), nullable=True, doc="""
+    Date and time when the cost became effective, if known.
+    """)
+
+    discount_starts = sa.Column(sa.DateTime(), nullable=True, doc="""
+    Date and time when the "current" discount for the cost begins, if
+    applicable.  Note that this should never be a "future" value, since only
+    the current discount should be part of this record.
+    """)
+
+    discount_ends = sa.Column(sa.DateTime(), nullable=True, doc="""
+    Date and time when the "current" discount for the cost ends, if applicable.
+    Note that this should never be a "past" value, since only the current
+    discount should be part of this record.
+    """)
+
+    discount_amount = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
+    Dollar amount of the "current" discount for the cost, if applicable.
+    """)
+
+    discount_percent = sa.Column(sa.Numeric(precision=7, scale=4), nullable=True, doc="""
+    Percentage amount of the "current" discount for the cost, if applicable.
+    """)
+
+    discount_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
+    Discounted (effective) unit cost for the item, while current discount is in
+    effect (if applicable).
+    """)
 
     discontinued = sa.Column(sa.Boolean(), nullable=True, doc="""
     Flag to indicate if the cost record has been discontinued, presumably by
@@ -509,6 +635,97 @@ Product.vendor = orm.relationship(
     viewonly=True)
 
 
+class ProductFutureCost(Base):
+    """
+    Represents a *future* cost record, i.e. an otherwise normal / complete cost
+    record, but which will become effective at some later date/time.
+
+    Note that this is meant to be more of a "queue" table in practice,
+    i.e. records which exist here should eventually be applied to the
+    :class:`ProductCost` table, at which point the future record(s) would be
+    deleted.
+    """
+    __tablename__ = 'product_future_cost'
+    __table_args__ = (
+        sa.ForeignKeyConstraint(['cost_uuid'], ['product_cost.uuid'], name='product_future_cost_fk_cost'),
+        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_future_cost_fk_product'),
+        sa.ForeignKeyConstraint(['vendor_uuid'], ['vendor.uuid'], name='product_future_cost_fk_vendor'),
+    )
+
+    uuid = uuid_column()
+
+    cost_uuid = sa.Column(sa.String(length=32), nullable=True)
+    cost = orm.relationship(
+        ProductCost,
+        doc="""
+        Reference to the "current" cost record to which this future cost record
+        will be applied (when the time comes), if applicable.  If this is
+        ``None``, then a new "current" cost record would be created instead.
+        """,
+        backref=orm.backref('futures'))
+
+    product_uuid = sa.Column(sa.String(length=32), nullable=False)
+    product = orm.relationship(
+        Product,
+        doc="""
+        Reference to the product to which the cost record pertains.
+        """,
+        backref=orm.backref(
+            'future_costs',
+            doc="""
+            Sequence of future cost records for the product, i.e. which have
+            yet to become "current" costs.
+            """))
+
+    vendor_uuid = sa.Column(sa.String(length=32), nullable=False)
+    vendor = orm.relationship(
+        Vendor,
+        doc="""
+        Reference to the vendor to which the cost record pertains.
+        """,
+        backref=orm.backref(
+            'future_costs',
+            doc="""
+            Sequence of future cost records for the vendor, i.e. which have yet
+            to become "current" costs.
+            """))
+
+    # TODO: we don't want this to be part of it, right? (at least not yet)
+    # preference = sa.Column(sa.Integer(), nullable=False)
+
+    order_code = sa.Column(sa.String(length=20), nullable=True, doc="""
+    Vendor order code for the item, if applicable.
+    """)
+
+    case_quantity = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
+    Number of units which constitute a "case" in the context of this vendor
+    catalog item.  May be a fractional quantity, e.g. 17.5LB.
+    """)
+
+    case_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
+    Case cost for the item, if known.
+    """)
+
+    unit_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
+    Unit cost for the item, if known.
+    """)
+
+    starts = sa.Column(sa.DateTime(), nullable=False, doc="""
+    Date and time when the cost becomes effective.
+    """)
+
+    ends = sa.Column(sa.DateTime(), nullable=True, doc="""
+    Date and time when the cost *stops* being effective, if applicable.  This
+    often will be null, in which case the cost becomes "permanently" effective,
+    i.e. until a newer cost is brought in.
+    """)
+
+    discontinued = sa.Column(sa.Boolean(), nullable=True, doc="""
+    Flag to indicate if the cost record has been discontinued, presumably by
+    the vendor.
+    """)
+
+
 class ProductPrice(Base):
     """
     Represents a price for a product.
@@ -552,6 +769,12 @@ Product.prices = orm.relationship(
     primaryjoin=ProductPrice.product_uuid == Product.uuid,
     cascade='save-update, merge, delete, delete-orphan')
 
+Product.suggested_price = orm.relationship(
+    ProductPrice,
+    primaryjoin=Product.suggested_price_uuid == ProductPrice.uuid,
+    lazy='joined',
+    post_update=True)
+
 Product.regular_price = orm.relationship(
     ProductPrice,
     primaryjoin=Product.regular_price_uuid == ProductPrice.uuid,
@@ -588,6 +811,7 @@ class ProductInventory(Base):
         backref=orm.backref(
             'inventory',
             uselist=False,
+            cascade='all, delete-orphan',
             doc="""
             Inventory data for the product, if any.
             """))
@@ -622,6 +846,7 @@ class ProductStoreInfo(Base):
         backref=orm.backref(
             'store_infos',
             collection_class=attribute_mapped_collection('store_uuid'),
+            cascade='all, delete-orphan',
             doc="""
             List of store-specific info records for the product.
             """))
@@ -637,3 +862,79 @@ class ProductStoreInfo(Base):
     Flag indicating the product has seen "recent activity" at the store.  How
     this is populated and/or interpreted is up to custom app logic.
     """)
+
+
+class ProductVolatile(Base):
+    """
+    This is the place to find "volatile" data for a given product, or at least
+    it should be...  As of this writing there are a couple other places to look
+    but hopefully this table can eventually be "the" place.
+
+    Whether any given value in a given record, applies to the "current" app
+    node only, or if it applies to all nodes, is up to app logic.
+
+    Note that a custom app should (most likely) *not* bother "extending" this
+    table, but rather should create a separate table with similar pattern.
+    """
+    __tablename__ = 'product_volatile'
+    __table_args__ = (
+        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_volatile_fk_product'),
+    )
+
+    uuid = uuid_column()
+
+    product_uuid = sa.Column(sa.String(length=32), nullable=False)
+    product = orm.relationship(
+        Product,
+        doc="""
+        Product to which this "volatile" data record pertains.
+        """,
+        backref=orm.backref(
+            'volatile',
+            uselist=False,
+            cascade='all, delete-orphan',
+            doc="""
+            "Volatile" data record for the product, if any.
+            """))
+
+    true_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
+    "True" unit cost for the item, if known.  This might include certain
+    "allowances" (discounts) currently in effect etc.; really anything which
+    might not be reflected in "official" unit cost for the product.  Usually,
+    this value is quite easily calculated and so this field serves as more of
+    a cache, for sake of SQL access to the values.
+    """)
+
+    true_margin = sa.Column(sa.Numeric(precision=8, scale=5), nullable=True, doc="""
+    "True" profit margin for the "regular" unit price vs. the "true" unit cost
+    (:attr:`true_cost`).
+    """)
+
+
+@six.python_2_unicode_compatible
+class InventoryAdjustmentReason(Base):
+    """
+    Reasons for adjusting product inventory.
+    """
+    __tablename__ = 'invadjust_reason'
+    __table_args__ = (
+        sa.UniqueConstraint('code', name='invadjust_reason_uq_code'),
+    )
+
+    uuid = uuid_column()
+
+    code = sa.Column(sa.String(length=20), nullable=False, doc="""
+    Unique code for the reason.
+    """)
+
+    description = sa.Column(sa.String(length=255), nullable=False, doc="""
+    Description for the reason.
+    """)
+
+    hidden = sa.Column(sa.Boolean(), nullable=True, doc="""
+    Flag indicating that the reason code should *not* be generally visible for
+    selection by the user etc.
+    """)
+
+    def __str__(self):
+        return self.description or ""
