@@ -1,8 +1,8 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; -*-
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2018 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -28,14 +28,17 @@ from __future__ import unicode_literals, absolute_import
 
 import datetime
 
+import six
 import sqlalchemy as sa
 from sqlalchemy import orm
 from sqlalchemy.ext.declarative import declared_attr
 
 from rattail.db.model import Base, uuid_column, Store, Department, Vendor, Employee, User, Product
 from rattail.db.types import GPCType
+from rattail.util import pretty_quantity
 
 
+@six.python_2_unicode_compatible
 class PurchaseBase(object):
     """
     Base class for purchases; defines common fields.
@@ -84,14 +87,15 @@ class PurchaseBase(object):
             Reference to the primary :class:`Department` for which the purchase was made.
             """)
 
-    buyer_uuid = sa.Column(sa.String(length=32), nullable=False)
+    buyer_uuid = sa.Column(sa.String(length=32), nullable=True)
 
     @declared_attr
     def buyer(cls):
         return orm.relationship(
             Employee,
             doc="""
-            Reference to the :class:`Employee` who placed the order with the vendor.
+            Reference to the :class:`Employee` who placed the order with the
+            vendor, if applicable/known.
             """)
 
     po_number = sa.Column(sa.String(length=20), nullable=True, doc="""
@@ -99,13 +103,21 @@ class PurchaseBase(object):
     """)
 
     po_total = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
-    Purchase order total, i.e. expected cost as of order placement.
+    Total cost according to the initial purchase order.
     """)
 
     date_ordered = sa.Column(sa.Date(), nullable=True, doc="""
     Date on which the purchase order was first submitted to the vendor.
     """)
 
+    ship_method = sa.Column(sa.String(length=50), nullable=True, doc="""
+    Code representing the shipping method.
+    """)
+
+    notes_to_vendor = sa.Column(sa.Text(), nullable=True, doc="""
+    Any arbitrary notes to the vendor, regarding the purchase.
+    """)
+
     date_shipped = sa.Column(sa.Date(), nullable=True, doc="""
     Date on which the order was shipped from the vendor.
     """)
@@ -123,14 +135,14 @@ class PurchaseBase(object):
     """)
 
     invoice_total = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
-    Invoice total, if applicable.
+    Total cost according to the invoice.
     """)
 
-    def __unicode__(self):
+    def __str__(self):
         if self.vendor and self.date_ordered:
             return "{} ({})".format(self.vendor, self.date_ordered.strftime('%Y-%m-%d'))
         if self.vendor:
-            return unicode(self.vendor)
+            return str(self.vendor)
         return ''
 
 
@@ -165,7 +177,12 @@ class PurchaseItemBase(object):
             Product,
             doc="""
             Reference to the :class:`Product` which the line item contains / represents.
-            """)
+            """,
+            backref=orm.backref(
+                '_{}_records'.format(cls.__tablename__),
+                doc="""
+                List of ``{}`` records associated with the product.
+                """.format(cls.__tablename__)))
 
     upc = sa.Column(GPCType(), nullable=True, doc="""
     Product UPC for the line item.
@@ -211,12 +228,25 @@ class PurchaseItemBase(object):
     Line number from the PO if known, for cross-reference.
     """)
 
+    catalog_unit_cost = sa.Column(sa.Numeric(precision=7, scale=3), nullable=True, doc="""
+    This corresponds to the ``ProductCost.unit_cost`` (aka. "catalog" or
+    wholesale cost) amount for the product/vendor, if applicable.
+    """)
+
     po_unit_cost = sa.Column(sa.Numeric(precision=7, scale=3), nullable=True, doc="""
     Expected cost per single unit of product, as of initial order placement.
     """)
 
     po_total = sa.Column(sa.Numeric(precision=7, scale=2), nullable=True, doc="""
-    Expected total cost for line item, as of initial order placement.
+    Total cost for the line item, according to the initial purchase order.
+    """)
+
+    cases_shipped = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
+    Number of cases of product which were supposedly shipped by/from the vendor.
+    """)
+
+    units_shipped = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
+    Number of units of product which were supposedly shipped by/from the vendor.
     """)
 
     cases_received = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
@@ -227,6 +257,11 @@ class PurchaseItemBase(object):
     Number of units of product which were ultimately received.
     """)
 
+    out_of_stock = sa.Column(sa.Boolean(), nullable=True, doc="""
+    Flag indicating whether the item was known to be "out of stock" per the
+    vendor invoice.
+    """)
+
     invoice_line_number = sa.Column(sa.Integer(), nullable=True, doc="""
     Line number from the invoice if known, for cross-reference.
     """)
@@ -240,7 +275,7 @@ class PurchaseItemBase(object):
     """)
 
     invoice_total = sa.Column(sa.Numeric(precision=7, scale=2), nullable=True, doc="""
-    Actual total cost for line item, per invoice.
+    Total cost for the line item, according to the invoice.
     """)
 
     cases_damaged = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
@@ -268,6 +303,7 @@ class PurchaseItemBase(object):
     """)
 
 
+@six.python_2_unicode_compatible
 class PurchaseCreditBase(object):
     """
     Base class for purchase credits.
@@ -345,6 +381,10 @@ class PurchaseCreditBase(object):
     Product UPC for the credit item.
     """)
 
+    vendor_item_code = sa.Column(sa.String(length=20), nullable=True, doc="""
+    Vendor-specific code for the credit item.
+    """)
+
     brand_name = sa.Column(sa.String(length=100), nullable=True, doc="""
     Brand name for the credit item.
     """)
@@ -401,6 +441,10 @@ class PurchaseCreditBase(object):
     Actual total cost for line item, per invoice.
     """)
 
+    credit_total = sa.Column(sa.Numeric(precision=7, scale=2), nullable=True, doc="""
+    Actual total cost for credit, i.e. value of missing/damaged product.
+    """)
+
     mispick_product_uuid = sa.Column(sa.String(length=32), nullable=True)
 
     @declared_attr
@@ -429,6 +473,26 @@ class PurchaseCreditBase(object):
     Product size for the mispick item.
     """)
 
+    def __str__(self):
+        if self.cases_shorted is not None and self.units_shorted is not None:
+            qty = "{} cases, {} units".format(
+                pretty_quantity(self.cases_shorted),
+                pretty_quantity(self.units_shorted))
+        elif self.cases_shorted is not None:
+            qty = "{} cases".format(pretty_quantity(self.cases_shorted))
+        elif self.units_shorted is not None:
+            qty = "{} units".format(pretty_quantity(self.units_shorted))
+        else:
+            qty = "??"
+        qty += " {}".format(self.credit_type)
+        if self.credit_type == 'expired' and self.expiration_date:
+            qty += " ({})".format(self.expiration_date)
+        if self.credit_type == 'mispick' and self.mispick_product:
+            qty += " ({})".format(self.mispick_product)
+        if self.invoice_total:
+            return "{} = ${:0.2f}".format(qty, self.invoice_total)
+        return qty
+
 
 class Purchase(PurchaseBase, Base):
     """
