@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2018 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -44,11 +44,14 @@ class Role(Base):
     __tablename__ = 'role'
     __table_args__ = (
         sa.UniqueConstraint('name', name='role_uq_name'),
-        )
+    )
     __versioned__ = {}
 
     uuid = uuid_column()
-    name = sa.Column(sa.String(length=25), nullable=False)
+
+    name = sa.Column(sa.String(length=100), nullable=False, doc="""
+    Name for the role.  Each role must have a name, which must be unique.
+    """)
 
     session_timeout = sa.Column(sa.Integer(), nullable=True, doc="""
     Optional session timeout value for the role, in seconds.  If this is set to
@@ -101,13 +104,20 @@ class User(Base):
         sa.ForeignKeyConstraint(['person_uuid'], ['person.uuid'], name='user_fk_person'),
         sa.UniqueConstraint('username', name='user_uq_username'),
         )
-    __versioned__ = {'exclude': ['password', 'salt']}
+    __versioned__ = {'exclude': ['password', 'salt', 'last_login']}
 
     uuid = uuid_column()
     username = sa.Column(sa.String(length=25), nullable=False)
     password = sa.Column(sa.String(length=60))
     salt = sa.Column(sa.String(length=29))
+
     person_uuid = sa.Column(sa.String(length=32))
+    person = orm.relationship(
+        Person,
+        uselist=False,
+        backref=orm.backref(
+            'user',
+            uselist=False))
 
     active = sa.Column(sa.Boolean(), nullable=False, default=True, doc="""
     Whether the user is active, e.g. allowed to log in via the UI.
@@ -122,6 +132,15 @@ class User(Base):
     as off-limits from the general active flag sync.
     """)
 
+    local_only = sa.Column(sa.Boolean(), nullable=True, doc="""
+    Flag indicating the user account is somehow specific to the "local" app
+    node etc. and should not be synced elsewhere.
+    """)
+
+    last_login = sa.Column(sa.DateTime(), nullable=True, doc="""
+    Timestamp when user last logged into the system.
+    """)
+
     def __str__(self):
         if self.person and str(self.person):
             return str(self.person)
@@ -206,15 +225,9 @@ class User(Base):
         self.events.append(UserEvent(**kwargs))
 
 
-User.person = orm.relationship(
-    Person,
-    back_populates='user',
-    uselist=False)
-
-Person.user = orm.relationship(
-    User,
-    back_populates='person',
-    uselist=False)
+Person.make_proxy(User, 'person', 'first_name')
+Person.make_proxy(User, 'person', 'last_name')
+Person.make_proxy(User, 'person', 'display_name')
 
 
 class UserRole(Base):
@@ -271,6 +284,7 @@ class UserEvent(Base):
         """,
         backref=orm.backref(
             'events',
+            cascade='all, delete-orphan',
             doc="""
             Sequence of events for the user.
             """))
