@@ -72,6 +72,19 @@ BATCH_ACTION = {
     }
 
 
+EMAIL_ATTEMPT_CREATED           = 0
+EMAIL_ATTEMPT_SUCCESS           = 1
+EMAIL_ATTEMPT_FAILURE           = 2
+# EMAIL_ATTEMPT_BOUNCED           = 3
+
+EMAIL_ATTEMPT = {
+    EMAIL_ATTEMPT_CREATED       : "created",
+    EMAIL_ATTEMPT_SUCCESS       : "success",
+    EMAIL_ATTEMPT_FAILURE       : "failure",
+    # EMAIL_ATTEMPT_BOUNCED       : "bounced",
+}
+
+
 EMAIL_PREFERENCE_NONE           = 0
 EMAIL_PREFERENCE_TEXT           = 1
 EMAIL_PREFERENCE_HTML           = 2
@@ -94,14 +107,31 @@ HANDHELD_DEVICE_TYPE = {
 }
 
 
+IMPORTER_BATCH_ROW_STATUS_NOCHANGE      = 0
+IMPORTER_BATCH_ROW_STATUS_CREATE        = 1
+IMPORTER_BATCH_ROW_STATUS_UPDATE        = 2
+IMPORTER_BATCH_ROW_STATUS_DELETE        = 3
+
+IMPORTER_BATCH_ROW_STATUS = {
+    IMPORTER_BATCH_ROW_STATUS_NOCHANGE  : "no change",
+    IMPORTER_BATCH_ROW_STATUS_CREATE    : "create",
+    IMPORTER_BATCH_ROW_STATUS_UPDATE    : "update",
+    IMPORTER_BATCH_ROW_STATUS_DELETE    : "delete",
+}
+
+
 INVENTORY_MODE_REPLACE          = 1
 INVENTORY_MODE_REPLACE_ADJUST   = 2
 INVENTORY_MODE_ADJUST           = 3
+INVENTORY_MODE_ZERO_ALL         = 4
+INVENTORY_MODE_VARIANCE         = 5
 
 INVENTORY_MODE = {
     INVENTORY_MODE_REPLACE              : "Replace only",
     INVENTORY_MODE_REPLACE_ADJUST       : "Replace then adjust",
     INVENTORY_MODE_ADJUST               : "Adjust only",
+    INVENTORY_MODE_ZERO_ALL             : "Zero all",
+    INVENTORY_MODE_VARIANCE             : "Variance Correction",
 }
 
 
@@ -155,6 +185,19 @@ PURCHASE_BATCH_MODE = {
 }
 
 
+PURCHASE_CREDIT_STATUS_NEW              = 10
+PURCHASE_CREDIT_STATUS_SUBMITTED        = 20
+PURCHASE_CREDIT_STATUS_SATISFIED        = 30
+PURCHASE_CREDIT_STATUS_NONCREDITABLE    = 40
+
+PURCHASE_CREDIT_STATUS = {
+    PURCHASE_CREDIT_STATUS_NEW          : "new",
+    PURCHASE_CREDIT_STATUS_SUBMITTED    : "submitted",
+    PURCHASE_CREDIT_STATUS_SATISFIED    : "satisfied",
+    PURCHASE_CREDIT_STATUS_NONCREDITABLE: "non-creditable",
+}
+
+
 PURCHASE_STATUS_NEW             = 1 # TODO: is this needed?
 PURCHASE_STATUS_ORDERED         = 10
 PURCHASE_STATUS_RECEIVED        = 20
@@ -179,18 +222,33 @@ TEMPMON_APPLIANCE_TYPE = {
 }
 
 
+TEMPMON_DISK_TYPE_SDCARD                = 1
+TEMPMON_DISK_TYPE_USB                   = 2
+
+TEMPMON_DISK_TYPE = {
+    TEMPMON_DISK_TYPE_SDCARD            : "SD card",
+    TEMPMON_DISK_TYPE_USB               : "USB",
+}
+
+
 TEMPMON_PROBE_STATUS_GOOD_TEMP          = 1
 TEMPMON_PROBE_STATUS_LOW_TEMP           = 2
 TEMPMON_PROBE_STATUS_HIGH_TEMP          = 3
-TEMPMON_PROBE_STATUS_CRITICAL_TEMP      = 4
+TEMPMON_PROBE_STATUS_CRITICAL_HIGH_TEMP = 4
+# TODO: deprecate / remove this one
+TEMPMON_PROBE_STATUS_CRITICAL_TEMP      = TEMPMON_PROBE_STATUS_CRITICAL_HIGH_TEMP
 TEMPMON_PROBE_STATUS_ERROR              = 5
+TEMPMON_PROBE_STATUS_CRITICAL_LOW_TEMP  = 6
 
 TEMPMON_PROBE_STATUS = {
-    TEMPMON_PROBE_STATUS_GOOD_TEMP      : "good temp",
-    TEMPMON_PROBE_STATUS_LOW_TEMP       : "low temp",
-    TEMPMON_PROBE_STATUS_HIGH_TEMP      : "high temp",
-    TEMPMON_PROBE_STATUS_CRITICAL_TEMP  : "critical temp",
-    TEMPMON_PROBE_STATUS_ERROR          : "error",
+    TEMPMON_PROBE_STATUS_CRITICAL_HIGH_TEMP     : "critical high temp",
+    # TODO: deprecate / remove this one
+    TEMPMON_PROBE_STATUS_CRITICAL_TEMP          : "critical high temp",
+    TEMPMON_PROBE_STATUS_HIGH_TEMP              : "high temp",
+    TEMPMON_PROBE_STATUS_GOOD_TEMP              : "good temp",
+    TEMPMON_PROBE_STATUS_LOW_TEMP               : "low temp",
+    TEMPMON_PROBE_STATUS_CRITICAL_LOW_TEMP      : "critical low temp",
+    TEMPMON_PROBE_STATUS_ERROR                  : "error",
 }
 
 
@@ -364,12 +422,29 @@ UNIT_OF_MEASURE = {
     }
 
 
+UPGRADE_STATUS_PENDING          = 1
+UPGRADE_STATUS_EXECUTING        = 2
+UPGRADE_STATUS_SUCCEEDED        = 3
+UPGRADE_STATUS_FAILED           = 4
+
+UPGRADE_STATUS = {
+    UPGRADE_STATUS_PENDING      : "pending execution",
+    UPGRADE_STATUS_EXECUTING    : "currently executing",
+    UPGRADE_STATUS_SUCCEEDED    : "execution succeeded",
+    UPGRADE_STATUS_FAILED       : "execution failed",
+}
+
+
 USER_EVENT_LOGIN                = 1
 USER_EVENT_LOGOUT               = 2
+USER_EVENT_BECOME_ROOT          = 3
+USER_EVENT_STOP_ROOT            = 4
 
 USER_EVENT = {
     USER_EVENT_LOGIN            : "login",
     USER_EVENT_LOGOUT           : "logout",
+    USER_EVENT_BECOME_ROOT      : "become root",
+    USER_EVENT_STOP_ROOT        : "stop being root",
 }
 
 
