@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2018 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -28,10 +28,13 @@ from __future__ import unicode_literals, absolute_import
 
 import logging
 
+import six
+
 from rattail.db import cache
 from rattail.db.util import QuerySequence
 from rattail.time import make_utc
-from rattail.util import progress_loop, OrderedDict
+from rattail.util import data_diffs, progress_loop, OrderedDict
+from rattail.csvutil import UnicodeDictReader
 
 
 log = logging.getLogger(__name__)
@@ -45,6 +48,17 @@ class ImportLimitReached(Exception):
 class Importer(object):
     """
     Base class for all data importers.
+
+    .. attribute:: direction
+
+       Should be a string, either ``'import'`` or ``'export'``.  This value is
+       used to improve verbiage for logging and other output, for a better
+       overall user experience.  It may also be used by importer logic, where
+       the direction would otherwise be ambiguous.
+
+       Note that the handler is responsible for assigning this value; the
+       importer should not define it.  See also
+       :attr:`rattail.importing.handlers.ImportHandler.direction`.
     """
     # Set this to the data model class which is targeted on the local side.
     model_class = None
@@ -82,14 +96,29 @@ class Importer(object):
     host_system_title = None
     local_system_title = None
 
-    def __init__(self, config=None, fields=None, key=None, **kwargs):
+    # TODO
+    # Whether or not the registered "importer" batch handler is able to handle
+    # batches for this importer (and/or, whether this importer is able to
+    # provide what's needed for the same).
+    batches_supported = False
+
+    # TODO
+    # If ``batches_supported`` is true, this should contain SQLAlchemy
+    # ``Column`` instance overrides, keyed by fieldname.  Any field not
+    # represented here will be given the default column type (string).
+    field_coltypes = {}
+
+    def __init__(self, config=None, key=None, direction='import', fields=None, exclude_fields=None, **kwargs):
         self.config = config
         self.enum = config.get_enum() if config else None
         self.model_class = kwargs.pop('model_class', self.get_model_class())
-        self.fields = fields or self.supported_fields
         if key is not None:
             self.key = key
-        if isinstance(self.key, basestring):
+        self.direction = direction
+        self.fields = fields or self.supported_fields
+        if exclude_fields:
+            self.exclude_fields(*exclude_fields)
+        if isinstance(self.key, six.string_types):
             self.key = (self.key,)
         if self.key:
             for field in self.key:
@@ -101,6 +130,24 @@ class Importer(object):
             self.model_name = self.model_class.__name__
         self._setup(**kwargs)
 
+    def include_fields(self, *args):
+        """
+        Add the given fields to the supported field list for the importer.  May
+        be used at runtime to customize behavior.
+        """
+        for field in args:
+            if field not in self.fields:
+                self.fields.append(field)
+
+    def exclude_fields(self, *args):
+        """
+        Remove the given fields from the supported field list for the importer.
+        May be used at runtime to customize behavior.
+        """
+        for field in args:
+            if field in self.fields:
+                self.fields.remove(field)
+
     def get_model_class(self):
         return self.model_class
 
@@ -118,7 +165,7 @@ class Importer(object):
         self.create = kwargs.pop('create', self.allow_create) and self.allow_create
         self.update = kwargs.pop('update', self.allow_update) and self.allow_update
         self.delete = kwargs.pop('delete', False) and self.allow_delete
-        for key, value in kwargs.iteritems():
+        for key, value in kwargs.items():
             setattr(self, key, value)
 
     def setup(self):
@@ -126,6 +173,11 @@ class Importer(object):
         Perform any setup necessary, e.g. cache lookups for existing data.
         """
 
+    def datasync_setup(self):
+        """
+        Perform any setup necessary, in the context of a datasync job.
+        """
+
     def teardown(self):
         """
         Perform any cleanup after import, if necessary.
@@ -135,6 +187,20 @@ class Importer(object):
         factory = factory or self.progress
         return progress_loop(func, items, factory, **kwargs)
 
+    def unique_data(self, host_data):
+        # Prune duplicate keys from host/source data.  This is for the sake of
+        # sanity since duplicates typically lead to a ping-pong effect, where a
+        # "clean" (change-less) import is impossible.
+        unique = OrderedDict()
+        for data in host_data:
+            key = self.get_key(data)
+            if key in unique:
+                log.warning("duplicate records detected from {} for key: {}".format(
+                    self.host_system_title, key))
+            else:
+                unique[key] = data
+        return list(unique.values()), unique
+
     def import_data(self, host_data=None, now=None, **kwargs):
         """
         Import some data!  This is the core body of logic for that, regardless
@@ -152,19 +218,7 @@ class Importer(object):
         # Get complete set of normalized host data.
         if host_data is None:
             host_data = self.normalize_host_data()
-
-        # Prune duplicate keys from host/source data.  This is for the sake of
-        # sanity since duplicates typically lead to a ping-pong effect, where a
-        # "clean" (change-less) import is impossible.
-        unique = OrderedDict()
-        for data in host_data:
-            key = self.get_key(data)
-            if key in unique:
-                log.warning("duplicate records detected from {} for key: {}".format(
-                    self.host_system_title, key))
-            else:
-                unique[key] = data
-        host_data = list(unique.itervalues())
+        host_data, unique = self.unique_data(host_data)
 
         # Cache local data if appropriate.
         if self.caches_local_data:
@@ -241,12 +295,74 @@ class Importer(object):
                 self.flush_create_update()
 
         try:
-            self.progress_loop(import_, data, message="Importing {} data".format(self.model_name))
+            self.progress_loop(import_, data, message="{}ing {} data to {}".format(
+                self.direction.capitalize(), self.model_name, self.handler.local_title))
         except ImportLimitReached:
             pass
         self.flush_create_update_final()
         return created, updated
 
+    # def _populate_create_update(self, row_table, data):
+    #     """
+    #     Populate create and/or update records for the given batch row table,
+    #     according to the given host data set.
+    #     """
+    #     created = []
+    #     updated = []
+    #     # count = len(data)
+    #     # if not count:
+    #     #     return created, updated
+
+    #     def record(host_data, i):
+            
+    #         # fetch local object, using key from host data
+    #         key = self.get_key(host_data)
+    #         local_object = self.get_local_object(key)
+
+    #         # if we have a local object, but its data differs from host, make an update record
+    #         if local_object and self.update:
+    #             local_data = self.normalize_local_object(local_object)
+    #             diffs = self.data_diffs(local_data, host_data)
+    #             if diffs:
+    #                 log.debug("fields '{}' differed for local data: {}, host data: {}".format(
+    #                     ','.join(diffs), local_data, host_data))
+    #                 local_object = self.update_object(local_object, host_data, local_data)
+    #                 updated.append((local_object, local_data, host_data))
+    #                 if self.max_update and len(updated) >= self.max_update:
+    #                     log.warning("max of {} *updated* records has been reached; stopping now".format(self.max_update))
+    #                     raise ImportLimitReached()
+    #                 if self.max_total and (len(created) + len(updated)) >= self.max_total:
+    #                     log.warning("max of {} *total changes* has been reached; stopping now".format(self.max_total))
+    #                     raise ImportLimitReached()
+
+    #         # if we did not yet have a local object, make a create record
+    #         elif not local_object and self.create:
+    #             local_object = self.create_object(key, host_data)
+    #             if local_object:
+    #                 log.debug("created new {} {}: {}".format(self.model_name, key, local_object))
+    #                 created.append((local_object, host_data))
+    #                 if self.caches_local_data and self.cached_local_data is not None:
+    #                     self.cached_local_data[key] = {'object': local_object, 'data': self.normalize_local_object(local_object)}
+    #                 if self.max_create and len(created) >= self.max_create:
+    #                     log.warning("max of {} *created* records has been reached; stopping now".format(self.max_create))
+    #                     raise ImportLimitReached()
+    #                 if self.max_total and (len(created) + len(updated)) >= self.max_total:
+    #                     log.warning("max of {} *total changes* has been reached; stopping now".format(self.max_total))
+    #                     raise ImportLimitReached()
+    #             else:
+    #                 log.debug("did NOT create new {} for key: {}".format(self.model_name, key))
+
+    #         # flush changes every so often
+    #         if not self.batch_size or (len(created) + len(updated)) % self.batch_size == 0:
+    #             self.flush_create_update()
+
+    #     try:
+    #         self.progress_loop(record, data, message="Importing {} data".format(self.model_name))
+    #     except ImportLimitReached:
+    #         pass
+    #     # self.flush_create_update_final()
+    #     return created, updated
+
     def flush_create_update(self):
         """
         Perform any steps necessary to "flush" the create/update changes which
@@ -280,8 +396,11 @@ class Importer(object):
                     log.warning("max of {} *total changes* has been reached; stopping now".format(self.max_total))
                     raise ImportLimitReached()
 
-        self.progress_loop(delete, sorted(deleting),
-                           message="Deleting {} data".format(self.model_name))
+        try:
+            self.progress_loop(delete, sorted(deleting),
+                               message="Deleting {} data".format(self.model_name))
+        except ImportLimitReached:
+            pass
 
         return deleted
 
@@ -312,20 +431,15 @@ class Importer(object):
         if host_objects is None:
             host_objects = self.get_host_objects()
         normalized = []
-        count = len(host_objects)
-        if count == 0:
-            return normalized
-        prog = None
-        if self.progress:
-            prog = self.progress("Reading {} data from {}".format(self.model_name, self.host_system_title), count)
-        for i, obj in enumerate(host_objects, 1):
+
+        def normalize(obj, i):
             data = self.normalize_host_object(obj)
             if data:
                 normalized.append(data)
-            if prog:
-                prog.update(i)
-        if prog:
-            prog.destroy()
+
+        self.progress_loop(normalize, host_objects,
+                           message="Reading {} data from {}".format(
+                               self.model_name, self.host_system_title))
         return normalized
 
     def normalize_host_object(self, obj):
@@ -341,6 +455,12 @@ class Importer(object):
         """
         raise NotImplementedError
 
+    def cache_local_message(self):
+        """
+        Must return a message to be used for progress when fetching "local" data.
+        """
+        return "Reading {} data from {}".format(self.model_name, self.handler.local_title)
+
     def get_cache_key(self, obj, normal):
         """
         Get the primary cache key for a given object and normalized data.
@@ -411,11 +531,7 @@ class Importer(object):
         Find all (relevant) fields which differ between the host and local data
         values for a given record.
         """
-        diffs = []
-        for field in self.fields:
-            if local_data[field] != host_data[field]:
-                diffs.append(field)
-        return diffs
+        return data_diffs(local_data, host_data, fields=self.fields)
 
     def make_object(self):
         """
@@ -488,6 +604,29 @@ class Importer(object):
                 data[field], data[field2] = data[field2], None
 
 
+class FromCSV(Importer):
+    """
+    Generic base class for importers whose data source is a CSV file.
+    """
+
+    def setup(self):
+        if not hasattr(self, 'source_data_path'):
+            if hasattr(self, 'args') and hasattr(self.args, 'source_csv'):
+                self.source_data_path = self.args.source_csv
+
+    def get_host_objects(self):
+        if six.PY3:
+            source_csv_file = open(self.source_data_path, 'rt', encoding='latin_1')
+            reader = UnicodeDictReader(source_csv_file)
+        else:
+            source_csv_file = open(self.source_data_path, 'rb')
+            reader = UnicodeDictReader(source_csv_file, 'latin_1')
+
+        objects = list(reader)
+        source_csv_file.close()
+        return objects
+
+
 class FromQuery(Importer):
     """
     Generic base class for importers whose raw external data source is a
@@ -512,9 +651,10 @@ class FromDjango(FromQuery):
     """
     Base class for importers using Django ORM on the host side.
     """
+    django_dbkey = 'default'
 
     def query(self):
-        return self.host_model_class.objects.all()
+        return self.host_model_class.objects.using(self.django_dbkey).all()
 
 
 class BatchImporter(Importer):
@@ -574,7 +714,8 @@ class BatchImporter(Importer):
                                 .filter(getattr(self.model_class, key).in_(keys))
             return self.cache_model(self.model_class, key=self.get_cache_key,
                                     query=query, query_options=self.cache_query_options(),
-                                    normalizer=self.normalize_cache_object)
+                                    normalizer=self.normalize_cache_object,
+                                    message=self.cache_local_message())
         return {} # empty cache
 
 
