@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2019 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -29,6 +29,8 @@ from __future__ import unicode_literals, absolute_import
 import datetime
 import logging
 
+import six
+import sqlalchemy as sa
 from sqlalchemy import orm
 from sqlalchemy.orm.exc import NoResultFound
 
@@ -46,12 +48,53 @@ class ToRattail(ToSQLAlchemy):
     Base class for all Rattail model importers.
     """
     key = 'uuid'
+    extension_attr = None
     extension_fields = []
 
     @property
     def simple_fields(self):
         return super(ToRattail, self).simple_fields + self.extension_fields
 
+    def cache_query_options(self):
+        options = super(ToRattail, self).cache_query_options() or []
+        # eager join extension table if applicable
+        if self.extension_attr and self.fields_active(self.extension_fields):
+            options.append(orm.joinedload(getattr(self.model_class, self.extension_attr)))
+        return options
+
+    def normalize_local_object(self, obj):
+        data = super(ToRattail, self).normalize_local_object(obj)
+        if data and self.extension_fields:
+            for field in self.extension_fields:
+                if field in self.fields:
+                    data[field] = getattr(obj, field)
+        return data
+
+    def update_object(self, obj, host_data, local_data=None):
+        obj = super(ToRattail, self).update_object(obj, host_data, local_data)
+        if obj and self.extension_fields:
+            for field in self.extension_fields:
+                if field in self.fields:
+                    if not local_data or field not in local_data or local_data[field] != host_data[field]:
+                        setattr(obj, field, host_data[field])
+        return obj
+
+
+class GlobalMixin(object):
+    """
+    Mixin for importers which should avoid "local only" records.
+    """
+
+    def cache_query(self):
+        query = super(GlobalMixin, self).cache_query()
+
+        # never include "local only" people
+        query = query.filter(sa.or_(
+            self.model_class.local_only == False,
+            self.model_class.local_only == None))
+
+        return query
+
 
 class PersonImporter(ToRattail):
     """
@@ -60,6 +103,13 @@ class PersonImporter(ToRattail):
     model_class = model.Person
 
 
+class GlobalPersonImporter(GlobalMixin, PersonImporter):
+    """
+    This is a customized version of the :class:`PersonImporter`, which simply
+    avoids "local only" person accounts.
+    """
+
+
 class PersonEmailAddressImporter(ToRattail):
     """
     Person email address data importer.
@@ -180,6 +230,13 @@ class PersonMailingAddressImporter(ToRattail):
     model_class = model.PersonMailingAddress
 
 
+class PersonNoteImporter(ToRattail):
+    """
+    Person notes data importer.
+    """
+    model_class = model.PersonNote
+
+
 class UserImporter(ToRattail):
     """
     User data importer.
@@ -216,6 +273,16 @@ class AdminUserImporter(UserImporter):
     def supported_fields(self):
         return super(AdminUserImporter, self).supported_fields + ['admin']
 
+    def setup(self):
+        super(AdminUserImporter, self).setup()
+        self.admin_uuid = auth.administrator_role(self.session).uuid
+
+    def cache_query_options(self):
+        options = super(AdminUserImporter, self).cache_query_options()
+        if 'admin' in self.fields:
+            options.append(orm.joinedload(model.User._roles))
+        return options
+
     def get_admin(self, session=None):
         from rattail.db import auth
         return auth.administrator_role(session or self.session)
@@ -223,7 +290,7 @@ class AdminUserImporter(UserImporter):
     def normalize_local_object(self, user):
         data = super(AdminUserImporter, self).normalize_local_object(user)
         if 'admin' in self.fields:
-            data['admin'] = self.get_admin() in user.roles
+            data['admin'] = self.admin_uuid in [r.role_uuid for r in user._roles]
         return data
 
     def update_object(self, user, data, local_data=None):
@@ -240,6 +307,13 @@ class AdminUserImporter(UserImporter):
             return user
 
 
+class GlobalUserImporter(GlobalMixin, UserImporter):
+    """
+    This is a customized version of the :class:`UserImporter`, which simply
+    avoids "local only" user accounts.
+    """
+
+
 class MessageImporter(ToRattail):
     """
     User message data importer.
@@ -456,21 +530,32 @@ class EmployeeImporter(ToRattail):
                 del person.customers[:]
 
         if 'phone_number' in self.fields:
+            phones = [phone for phone in employee.phones if phone.type == 'Home']
             number = data['phone_number']
             if number:
-                found = False
-                for phone in employee.phones:
-                    if phone.type == 'Home':
-                        if phone.number != number:
-                            phone.number = number
-                        found = True
-                        break
-                if not found:
+                if phones:
+                    phone = phones[0]
+                    if phone.number != number:
+                        phone.number = number
+                else:
                     employee.add_phone_number(number, type='Home')
             else:
-                for phone in list(employee.phones):
-                    if phone.type == 'Home':
-                        employee.phones.remove(phone)
+                for phone in phones:
+                    employee.phones.remove(phone)
+
+        if 'phone_number_2' in self.fields:
+            phones = [phone for phone in employee.phones if phone.type == 'Home']
+            number = data['phone_number_2']
+            if number:
+                if len(phones) > 1:
+                    phone = phones[1]
+                    if phone.number != number:
+                        phone.number = number
+                else:
+                    employee.add_phone_number(number, 'Home')
+            else:
+                for phone in phones[1:]:
+                    employee.phones.remove(phone)
 
         if 'email_address' in self.fields:
             address = data['email_address']
@@ -514,6 +599,13 @@ class EmployeePhoneNumberImporter(ToRattail):
     model_class = model.EmployeePhoneNumber
 
 
+class EmployeeHistoryImporter(ToRattail):
+    """
+    Employee history data importer
+    """
+    model_class = model.EmployeeHistory
+
+
 class ScheduledShiftImporter(ToRattail):
     """
     Imports employee scheduled shifts.
@@ -784,6 +876,27 @@ class CustomerPhoneNumberImporter(ToRattail):
     model_class = model.CustomerPhoneNumber
 
 
+class MemberImporter(ToRattail):
+    """
+    Member data importer.
+    """
+    model_class = model.Member
+
+
+class MemberEmailAddressImporter(ToRattail):
+    """
+    Member email address data importer.
+    """
+    model_class = model.MemberEmailAddress
+
+
+class MemberPhoneNumberImporter(ToRattail):
+    """
+    Member phone number data importer.
+    """
+    model_class = model.MemberPhoneNumber
+
+
 class VendorImporter(ToRattail):
     """
     Vendor data importer.
@@ -808,6 +921,12 @@ class VendorImporter(ToRattail):
             'email_address',
         ]
 
+    def cache_query(self):
+        query = super(VendorImporter, self).cache_query()
+        if len(self.key) == 1 and self.key[0] == 'id':
+            query = query.filter(model.Vendor.id != None)
+        return query
+
     def cache_query_options(self):
         options = []
         if self.fields_active(self.phone_fields + self.fax_fields):
@@ -1054,6 +1173,12 @@ class CategoryImporter(ToRattail):
     """
     model_class = model.Category
 
+    # set to True in order to auto-create record for "unknown" department
+    auto_create_unknown_department = False
+
+    # set to False in order to log debug instead of warning for "unknown" department
+    warn_for_unknown_department = True
+
     @property
     def supported_fields(self):
         return self.simple_fields + [
@@ -1092,14 +1217,21 @@ class CategoryImporter(ToRattail):
             number = data['department_number']
             if number:
                 dept = self.get_department(number)
-                if not dept:
+                if dept:
+                    category.department = dept
+                elif self.auto_create_unknown_department:
                     dept = model.Department()
                     dept.number = number
                     dept.name = "(created from import)"
                     self.session.add(dept)
                     if hasattr(self, 'departments'):
                         self.departments[dept.number] = dept
-                category.department = dept
+                    category.department = dept
+                else:
+                    logger = log.warning if self.warn_for_unknown_department else log.debug
+                    logger("unknown department number '%s' for %s", number, category)
+                    if category.department:
+                        category.department = None
             elif category.department:
                 category.department = None
 
@@ -1134,6 +1266,13 @@ class TaxImporter(ToRattail):
     model_class = model.Tax
 
 
+class InventoryAdjustmentReasonImporter(ToRattail):
+    """
+    InventoryAdjustmentReason data importer.
+    """
+    model_class = model.InventoryAdjustmentReason
+
+
 class BrandImporter(ToRattail):
     """
     Brand data importer.
@@ -1157,6 +1296,17 @@ class ProductImporter(ToRattail):
         'regular_price_starts',
         'regular_price_ends',
     ]
+    tpr_price_fields = [
+        'tpr_price_price',
+        'tpr_price_multiple',
+        'tpr_price_pack_price',
+        'tpr_price_pack_multiple',
+        'tpr_price_type',
+        'tpr_price_level',
+        'tpr_price_starts',
+        'tpr_price_ends',
+        'tpr_price_current',
+    ]
     sale_price_fields = [
         'sale_price_price',
         'sale_price_multiple',
@@ -1166,6 +1316,17 @@ class ProductImporter(ToRattail):
         'sale_price_level',
         'sale_price_starts',
         'sale_price_ends',
+        'sale_price_current',
+    ]
+    suggested_price_fields = [
+        'suggested_price_price',
+        'suggested_price_multiple',
+        'suggested_price_pack_price',
+        'suggested_price_pack_multiple',
+        'suggested_price_type',
+        'suggested_price_level',
+        'suggested_price_starts',
+        'suggested_price_ends',
     ]
     vendor_fields = [
         'vendor_id',
@@ -1174,13 +1335,30 @@ class ProductImporter(ToRattail):
         'vendor_unit_cost',
     ]
 
-    extension_attr = None
-
     maxlen_category_code = maxlen(model.Category.code)
 
+    # set to True in order to auto-create record for "unknown" category
+    auto_create_unknown_category = False
+
+    # set to True in order to auto-create record for "unknown" report code
+    auto_create_unknown_report_code = False
+
+    # set to False in order to log debug instead of warning for "unknown" category
+    warn_for_unknown_category = True
+
+    # set to False in order to log debug instead of warning for "unknown" report code
+    warn_for_unknown_report_code = True
+
     @property
     def supported_fields(self):
-        return self.simple_fields + self.regular_price_fields + self.sale_price_fields + self.vendor_fields + [
+        return (
+            self.simple_fields
+            + self.regular_price_fields
+            + self.tpr_price_fields
+            + self.sale_price_fields
+            + self.suggested_price_fields
+            + self.vendor_fields
+        ) + [
             'brand_name',
             'department_number',
             'subdepartment_number',
@@ -1218,8 +1396,17 @@ class ProductImporter(ToRattail):
         if 'vendor_id' in self.fields:
             self.vendors = self.cache_model(model.Vendor, key='id')
 
+    def cache_query(self):
+        query = super(ProductImporter, self).cache_query()
+        if len(self.key) == 1:
+            if self.key[0] == 'item_id':
+                query = query.filter(model.Product.item_id != None)
+            elif self.key[0] == 'upc':
+                query = query.filter(model.Product.upc != None)
+        return query
+
     def cache_query_options(self):
-        options = []
+        options = super(ProductImporter, self).cache_query_options()
         if 'brand_name' in self.fields:
             options.append(orm.joinedload(model.Product.brand))
         if 'department_number' in self.fields:
@@ -1237,23 +1424,27 @@ class ProductImporter(ToRattail):
         if 'tax_code' in self.fields:
             options.append(orm.joinedload(model.Product.tax))
         joined_prices = False
-        if self.fields_active(self.regular_price_fields):
+        if self.fields_active(self.suggested_price_fields):
             options.append(orm.joinedload(model.Product.prices))
-            options.append(orm.joinedload(model.Product.regular_price))
             joined_prices = True
-        if self.fields_active(self.sale_price_fields):
+            options.append(orm.joinedload(model.Product.suggested_price))
+        if self.fields_active(self.regular_price_fields):
             if not joined_prices:
                 options.append(orm.joinedload(model.Product.prices))
+                joined_prices = True
+            options.append(orm.joinedload(model.Product.regular_price))
+        if self.fields_active(self.tpr_price_fields + self.sale_price_fields):
+            if not joined_prices:
+                options.append(orm.joinedload(model.Product.prices))
+                joined_prices = True
             options.append(orm.joinedload(model.Product.current_price))
+            if self.fields_active(self.tpr_price_fields):
+                options.append(orm.joinedload(model.Product.tpr_price))
+            if self.fields_active(self.sale_price_fields):
+                options.append(orm.joinedload(model.Product.sale_price))
         if self.fields_active(self.vendor_fields):
             options.append(orm.joinedload(model.Product.cost))
             # options.append(orm.joinedload(model.Product.costs))
-
-        # eager join extension table if applicable
-        if self.extension_attr and self.fields_active(self.extension_fields):
-            cls, name = self.extension_attr
-            options.append(orm.joinedload(getattr(cls, name)))
-
         return options
 
     def normalize_local_object(self, product):
@@ -1278,6 +1469,25 @@ class ProductImporter(ToRattail):
         if 'tax_code' in self.fields:
             data['tax_code'] = product.tax.code if product.tax else None
 
+        if self.fields_active(self.suggested_price_fields):
+            price = product.suggested_price
+            if 'suggested_price_price' in self.fields:
+                data['suggested_price_price'] = price.price if price else None
+            if 'suggested_price_multiple' in self.fields:
+                data['suggested_price_multiple'] = price.multiple if price else None
+            if 'suggested_price_pack_price' in self.fields:
+                data['suggested_price_pack_price'] = price.pack_price if price else None
+            if 'suggested_price_pack_multiple' in self.fields:
+                data['suggested_price_pack_multiple'] = price.pack_multiple if price else None
+            if 'suggested_price_type' in self.fields:
+                data['suggested_price_type'] = price.type if price else None
+            if 'suggested_price_level' in self.fields:
+                data['suggested_price_level'] = price.level if price else None
+            if 'suggested_price_starts' in self.fields:
+                data['suggested_price_starts'] = price.starts if price else None
+            if 'suggested_price_ends' in self.fields:
+                data['suggested_price_ends'] = price.ends if price else None
+
         if self.fields_active(self.regular_price_fields):
             price = product.regular_price
             if 'regular_price_price' in self.fields:
@@ -1297,8 +1507,29 @@ class ProductImporter(ToRattail):
             if 'regular_price_ends' in self.fields:
                 data['regular_price_ends'] = price.ends if price else None
 
+        if self.fields_active(self.tpr_price_fields):
+            price = product.tpr_price
+            if 'tpr_price_price' in self.fields:
+                data['tpr_price_price'] = price.price if price else None
+            if 'tpr_price_multiple' in self.fields:
+                data['tpr_price_multiple'] = price.multiple if price else None
+            if 'tpr_price_pack_price' in self.fields:
+                data['tpr_price_pack_price'] = price.pack_price if price else None
+            if 'tpr_price_pack_multiple' in self.fields:
+                data['tpr_price_pack_multiple'] = price.pack_multiple if price else None
+            if 'tpr_price_type' in self.fields:
+                data['tpr_price_type'] = price.type if price else None
+            if 'tpr_price_level' in self.fields:
+                data['tpr_price_level'] = price.level if price else None
+            if 'tpr_price_starts' in self.fields:
+                data['tpr_price_starts'] = price.starts if price else None
+            if 'tpr_price_ends' in self.fields:
+                data['tpr_price_ends'] = price.ends if price else None
+            if 'tpr_price_current' in self.fields:
+                data['tpr_price_current'] = bool(price and product.current_price is price)
+
         if self.fields_active(self.sale_price_fields):
-            price = product.current_price
+            price = product.sale_price
             if 'sale_price_price' in self.fields:
                 data['sale_price_price'] = price.price if price else None
             if 'sale_price_multiple' in self.fields:
@@ -1315,6 +1546,8 @@ class ProductImporter(ToRattail):
                 data['sale_price_starts'] = price.starts if price else None
             if 'sale_price_ends' in self.fields:
                 data['sale_price_ends'] = price.ends if price else None
+            if 'sale_price_current' in self.fields:
+                data['sale_price_current'] = bool(price and product.current_price is price)
 
         if self.fields_active(self.vendor_fields):
             cost = product.cost
@@ -1448,7 +1681,9 @@ class ProductImporter(ToRattail):
                     code = code[:self.maxlen_category_code]
 
                 category = self.get_category_by_code(code)
-                if not category:
+                if category:
+                    product.category = category
+                elif self.auto_create_unknown_category:
                     category = model.Category()
                     category.code = code
                     try:
@@ -1458,22 +1693,36 @@ class ProductImporter(ToRattail):
                     self.session.add(category)
                     if hasattr(self, 'categories'):
                         self.categories[code] = category
-                product.category = category
+                    product.category = category
+                else:
+                    logger = log.warning if self.warn_for_unknown_category else log.debug
+                    logger("unknown category code '%s' for %s", code, product)
+                    if product.category:
+                        product.category = None
+
             elif product.category:
                 product.category = None
 
         elif 'category_number' in self.fields:
             number = data['category_number']
             if number:
-                cat = self.get_category_by_number(number)
-                if not cat:
-                    cat = model.Category()
-                    cat.number = number
-                    cat.name = "(auto-created)"
-                    self.session.add(cat)
+                category = self.get_category_by_number(number)
+                if category:
+                    product.category = category
+                elif self.auto_create_unknown_category:
+                    category = model.Category()
+                    category.number = number
+                    category.code = six.text_type(number)
+                    category.name = "(auto-created)"
+                    self.session.add(category)
                     if hasattr(self, 'categories'):
-                        self.categories[number] = cat
-                product.category = cat
+                        self.categories[number] = category
+                    product.category = category
+                else:
+                    logger = log.warning if self.warn_for_unknown_category else log.debug
+                    logger("unknown category number '%s' for %s", number, product)
+                    if product.category:
+                        product.category = None
             elif product.category:
                 product.category = None
 
@@ -1496,14 +1745,21 @@ class ProductImporter(ToRattail):
             code = data['report_code']
             if code:
                 rc = self.get_reportcode(code)
-                if not rc:
+                if rc:
+                    product.report_code = rc
+                elif self.auto_create_unknown_report_code:
                     rc = model.ReportCode()
                     rc.code = code
-                    rc.name = "(auto-created)"
+                    rc.name = "(no name)"
                     self.session.add(rc)
                     if hasattr(self, 'reportcodes'):
                         self.reportcodes[rc.code] = rc
-                product.report_code = rc
+                    product.report_code = rc
+                else:
+                    logger = log.warning if self.warn_for_unknown_report_code else log.debug
+                    logger("unknown report code '%s' for %s", code, product)
+                    if product.report_code:
+                        product.report_code = None
             elif product.report_code:
                 product.report_code = None
 
@@ -1538,6 +1794,37 @@ class ProductImporter(ToRattail):
             elif product.tax:
                 product.tax = None
 
+        create = False
+        delete = False
+        if self.fields_active(self.suggested_price_fields):
+            delete = True
+            create = any([data.get(f) is not None for f in self.suggested_price_fields])
+        if create:
+            price = product.suggested_price
+            if not price:
+                price = model.ProductPrice()
+                price.type = self.enum.PRICE_TYPE_MFR_SUGGESTED
+                product.prices.append(price)
+                product.suggested_price = price
+            if 'suggested_price_price' in self.fields:
+                price.price = data['suggested_price_price']
+            if 'suggested_price_multiple' in self.fields:
+                price.multiple = data['suggested_price_multiple']
+            if 'suggested_price_pack_price' in self.fields:
+                price.pack_price = data['suggested_price_pack_price']
+            if 'suggested_price_pack_multiple' in self.fields:
+                price.pack_multiple = data['suggested_price_pack_multiple']
+            if 'suggested_price_type' in self.fields:
+                price.type = data['suggested_price_type']
+            if 'suggested_price_level' in self.fields:
+                price.level = data['suggested_price_level']
+            if 'suggested_price_starts' in self.fields:
+                price.starts = data['suggested_price_starts']
+            if 'suggested_price_ends' in self.fields:
+                price.ends = data['suggested_price_ends']
+        elif delete and product.suggested_price:
+            product.suggested_price = None
+
         create = False
         delete = False
         if self.fields_active(self.regular_price_fields):
@@ -1547,6 +1834,7 @@ class ProductImporter(ToRattail):
             price = product.regular_price
             if not price:
                 price = model.ProductPrice()
+                price.type = self.enum.PRICE_TYPE_REGULAR
                 product.prices.append(price)
                 product.regular_price = price
             if 'regular_price_price' in self.fields:
@@ -1570,15 +1858,54 @@ class ProductImporter(ToRattail):
 
         create = False
         delete = False
-        if self.fields_active(self.sale_price_fields):
+        if self.fields_active(self.tpr_price_fields):
             delete = True
-            create = any([data.get(f) is not None for f in self.sale_price_fields])
+            true_fields = set(self.tpr_price_fields) - set(['tpr_price_current'])
+            create = any([data.get(f) is not None for f in true_fields])
         if create:
-            price = product.current_price
+            price = product.tpr_price
             if not price:
                 price = model.ProductPrice()
+                price.type = self.enum.PRICE_TYPE_TPR
                 product.prices.append(price)
-                product.current_price = price
+                product.tpr_price = price
+            if 'tpr_price_price' in self.fields:
+                price.price = data['tpr_price_price']
+            if 'tpr_price_multiple' in self.fields:
+                price.multiple = data['tpr_price_multiple']
+            if 'tpr_price_pack_price' in self.fields:
+                price.pack_price = data['tpr_price_pack_price']
+            if 'tpr_price_pack_multiple' in self.fields:
+                price.pack_multiple = data['tpr_price_pack_multiple']
+            if 'tpr_price_type' in self.fields:
+                price.type = data['tpr_price_type']
+            if 'tpr_price_level' in self.fields:
+                price.level = data['tpr_price_level']
+            if 'tpr_price_starts' in self.fields:
+                price.starts = data['tpr_price_starts']
+            if 'tpr_price_ends' in self.fields:
+                price.ends = data['tpr_price_ends']
+            if 'tpr_price_current' in self.fields:
+                if data['tpr_price_current'] and product.current_price is not price:
+                    product.current_price = price
+        elif delete and product.tpr_price:
+            if product.current_price is product.tpr_price:
+                product.current_price = None
+            product.tpr_price = None
+
+        create = False
+        delete = False
+        if self.fields_active(self.sale_price_fields):
+            delete = True
+            true_fields = set(self.sale_price_fields) - set(['sale_price_current'])
+            create = any([data.get(f) is not None for f in true_fields])
+        if create:
+            price = product.sale_price
+            if not price:
+                price = model.ProductPrice()
+                price.type = self.enum.PRICE_TYPE_SALE
+                product.prices.append(price)
+                product.sale_price = price
             if 'sale_price_price' in self.fields:
                 price.price = data['sale_price_price']
             if 'sale_price_multiple' in self.fields:
@@ -1595,8 +1922,13 @@ class ProductImporter(ToRattail):
                 price.starts = data['sale_price_starts']
             if 'sale_price_ends' in self.fields:
                 price.ends = data['sale_price_ends']
-        elif delete and product.current_price:
-            product.current_price = None
+            if 'sale_price_current' in self.fields:
+                if data['sale_price_current'] and product.current_price is not price:
+                    product.current_price = price
+        elif delete and product.sale_price:
+            if product.current_price is product.sale_price:
+                product.current_price = None
+            product.sale_price = None
 
         if 'vendor_id' in self.fields:
             vendor_id = data['vendor_id']
@@ -1617,7 +1949,8 @@ class ProductImporter(ToRattail):
                 else:
                     cost = model.ProductCost()
                     cost.vendor = vendor
-                    product.costs.append(cost)
+                    with self.session.no_autoflush:
+                        product.costs.append(cost)
                     # TODO: This seems heavy-handed, but also seems necessary
                     # to populate the `Product.cost` relationship...
                     self.session.add(product)
@@ -1771,23 +2104,36 @@ class ProductCostImporter(ToRattail):
     """
     model_class = model.ProductCost
 
+    product_key_fields = [
+        'product_upc',
+        'product_item_id',
+    ]
+
     @property
     def supported_fields(self):
-        return self.simple_fields + [
-            'product_upc',
+        return self.simple_fields + self.product_key_fields + [
             'vendor_id',
             'preferred',
         ]
 
     def setup(self):
+
+        # note that only one product key field should be supported
         if 'product_upc' in self.fields:
-            self.products = self.cache_model(model.Product, key='upc')
+            query = self.session.query(model.Product)\
+                                .filter(model.Product.upc != None)
+            self.products = self.cache_model(model.Product, query=query, key='upc')
+        elif 'product_item_id' in self.fields:
+            query = self.session.query(model.Product)\
+                                .filter(model.Product.item_id != None)
+            self.products = self.cache_model(model.Product, query=query, key='item_id')
+
         if 'vendor_id' in self.fields:
             self.vendors = self.cache_model(model.Vendor, key='id')
 
     def cache_query_options(self):
         options = []
-        if 'product_upc' in self.fields:
+        if self.fields_active(self.product_key_fields):
             options.append(orm.joinedload(model.ProductCost.product))
         if 'vendor_id' in self.fields:
             options.append(orm.joinedload(model.ProductCost.vendor))
@@ -1796,12 +2142,11 @@ class ProductCostImporter(ToRattail):
     def get_single_local_object(self, key):
         """
         Fetch a single ``ProductCost`` object from local Rattail, taking
-        complex key fields (namely ``product_upc``, ``vendor_id``) into
-        account.
+        complex key fields (e.g. ``product_upc``, ``vendor_id``) into account.
         """
         query = self.session.query(model.ProductCost)
         for i, k in enumerate(self.key):
-            if k not in ('product_upc', 'vendor_id'):
+            if k not in ('product_upc', 'product_item_id', 'vendor_id'):
                 query = query.filter(getattr(self.model_class, k) == key[i])
 
         try:
@@ -1811,6 +2156,13 @@ class ProductCostImporter(ToRattail):
         else:
             query = query.join(model.Product).filter(model.Product.upc == key[i])
 
+        try:
+            i = self.key.index('product_item_id')
+        except ValueError:
+            pass
+        else:
+            query = query.join(model.Product).filter(model.Product.item_id == key[i])
+
         try:
             i = self.key.index('vendor_id')
         except ValueError:
@@ -1827,6 +2179,8 @@ class ProductCostImporter(ToRattail):
         data = super(ProductCostImporter, self).normalize_local_object(cost)
         if 'product_upc' in self.fields:
             data['product_upc'] = cost.product.upc
+        if 'product_item_id' in self.fields:
+            data['product_item_id'] = cost.product.item_id
         if 'vendor_id' in self.fields:
             data['vendor_id'] = cost.vendor.id
         if 'preferred' in self.fields:
@@ -1840,6 +2194,13 @@ class ProductCostImporter(ToRattail):
                            .filter(model.Product.upc == upc)\
                            .first()
 
+    def get_product_by_item_id(self, item_id):
+        if hasattr(self, 'products'):
+            return self.products.get(item_id)
+        return self.session.query(model.Product)\
+                           .filter(model.Product.item_id == item_id)\
+                           .first()
+
     def get_vendor(self, vendor_id):
         if hasattr(self, 'vendors'):
             return self.vendors.get(vendor_id)
@@ -1878,6 +2239,22 @@ class ProductCostImporter(ToRattail):
             elif cost.product is not product:
                 log.warning("duplicate products detected for UPC {}".format(upc.pretty()))
 
+        if 'product_item_id' in self.fields and 'product_uuid' not in self.fields:
+            item_id = data['product_item_id']
+            if not item_id:
+                raise ValueError("Source data has no product_item_id value: {}".format(data))
+            product = self.get_product_by_item_id(item_id)
+            if not product:
+                product = model.Product()
+                product.item_id = item_id
+                self.session.add(product)
+                if hasattr(self, 'products'):
+                    self.products[product.item_id] = product
+            if not cost.product:
+                product.costs.append(cost)
+            elif cost.product is not product:
+                log.warning("duplicate products detected for item_id: %s", item_id)
+
         if 'preferred' in self.fields:
             product = cost.product or self.session.query(model.Product).get(cost.product_uuid)
             if data['preferred']:
@@ -1913,6 +2290,96 @@ class ProductPriceImporter(ToRattail):
     """
     model_class = model.ProductPrice
 
+    product_reference_fields = [
+        'product_suggested_price',
+        'product_regular_price',
+        'product_tpr_price',
+        'product_sale_price',
+        'product_current_price',
+    ]
+
+    @property
+    def supported_fields(self):
+        return self.simple_fields + self.product_reference_fields
+
+    def cache_query_options(self):
+        options = []
+        if self.fields_active(self.product_reference_fields):
+            options.append(orm.joinedload(model.ProductPrice.product))
+        return options
+
+    def normalize_local_object(self, price):
+        data = super(ProductPriceImporter, self).normalize_local_object(price)
+
+        if self.fields_active(self.product_reference_fields):
+            product = price.product
+            if 'product_suggested_price' in self.fields:
+                data['product_suggested_price'] = bool(product.suggested_price is price)
+            if 'product_regular_price' in self.fields:
+                data['product_regular_price'] = bool(product.regular_price is price)
+            if 'product_tpr_price' in self.fields:
+                data['product_tpr_price'] = bool(product.tpr_price is price)
+            if 'product_sale_price' in self.fields:
+                data['product_sale_price'] = bool(product.sale_price is price)
+            if 'product_current_price' in self.fields:
+                data['product_current_price'] = bool(product.current_price is price)
+
+        return data
+
+    def update_object(self, price, data, local_data=None):
+        price = super(ProductPriceImporter, self).update_object(price, data, local_data)
+
+        if self.fields_active(self.product_reference_fields):
+            if not price.product:
+                self.session.add(price)
+                self.session.flush()
+                self.session.refresh(price)
+            product = price.product
+            assert product
+            assert price in product.prices
+
+            if 'product_suggested_price' in self.fields:
+                if data['product_suggested_price']:
+                    if product.suggested_price is not price:
+                        product.suggested_price = price
+                else:
+                    if product.suggested_price is price:
+                        product.suggested_price = None
+
+            if 'product_regular_price' in self.fields:
+                if data['product_regular_price']:
+                    if product.regular_price is not price:
+                        product.regular_price = price
+                else:
+                    if product.regular_price is price:
+                        product.regular_price = None
+
+            if 'product_tpr_price' in self.fields:
+                if data['product_tpr_price']:
+                    if product.tpr_price is not price:
+                        product.tpr_price = price
+                else:
+                    if product.tpr_price is price:
+                        product.tpr_price = None
+
+            if 'product_sale_price' in self.fields:
+                if data['product_sale_price']:
+                    if product.sale_price is not price:
+                        product.sale_price = price
+                else:
+                    if product.sale_price is price:
+                        product.sale_price = None
+
+            if 'product_current_price' in self.fields:
+                if data['product_current_price']:
+                    if product.current_price is not price:
+                        product.current_price = price
+                else:
+                    if product.current_price is price:
+                        product.current_price = None
+
+        return price
+
 
 class ProductStoreInfoImporter(ToRattail):
     """
@@ -1921,6 +2388,29 @@ class ProductStoreInfoImporter(ToRattail):
     model_class = model.ProductStoreInfo
 
 
+class ProductVolatileImporter(ToRattail):
+    """
+    Data importer for :class:`~rattail.db.model.products.ProductVolatile`.
+    """
+    model_class = model.ProductVolatile
+
+
+class LabelProfileImporter(ToRattail):
+    """
+    Importer for LabelProfile data
+    """
+    model_class = model.LabelProfile
+
+    def cache_query(self):
+        query = super(LabelProfileImporter, self).cache_query()
+
+        if not self.config.getbool('rattail', 'labels.sync_all_profiles', default=False):
+            # only fetch labels from host which are marked as "sync me"
+            query = query .filter(self.model_class.sync_me == True)
+
+        return query
+
+
 class CustomerOrderImporter(ToRattail):
     """
     Importer for CustomerOrder data
