@@ -47,6 +47,7 @@ class FromRattailToRattailVersions(FromRattailHandler, ToRattailHandler):
         importers['PersonPhoneNumber'] = PersonPhoneNumberImporter
         importers['PersonEmailAddress'] = PersonEmailAddressImporter
         importers['PersonMailingAddress'] = PersonMailingAddressImporter
+        importers['PersonNote'] = PersonNoteImporter
         importers['Role'] = RoleImporter
         # importers['Permission'] = PermissionImporter
         importers['User'] = UserImporter
@@ -66,6 +67,7 @@ class FromRattailToRattailVersions(FromRattailHandler, ToRattailHandler):
         importers['EmployeeEmailAddress'] = EmployeeEmailAddressImporter
         importers['EmployeeStore'] = EmployeeStoreImporter
         importers['EmployeeDepartment'] = EmployeeDepartmentImporter
+        importers['EmployeeHistory'] = EmployeeHistoryImporter
         importers['WorkedShift'] = WorkedShiftImporter
         importers['Vendor'] = VendorImporter
         importers['VendorPhoneNumber'] = VendorPhoneNumberImporter
@@ -164,6 +166,9 @@ class PersonEmailAddressImporter(VersionImporter):
 class PersonMailingAddressImporter(VersionImporter):
     host_model_class = model.PersonMailingAddress
 
+class PersonNoteImporter(VersionImporter):
+    host_model_class = model.PersonNote
+
 class RoleImporter(VersionImporter):
     host_model_class = model.Role
 
@@ -221,6 +226,9 @@ class EmployeeStoreImporter(VersionImporter):
 class EmployeeDepartmentImporter(VersionImporter):
     host_model_class = model.EmployeeDepartment
 
+class EmployeeHistoryImporter(VersionImporter):
+    host_model_class = model.EmployeeHistory
+
 class WorkedShiftImporter(VersionImporter):
     host_model_class = model.WorkedShift
 
