@@ -1,6 +1,6 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; -*-
 
-from __future__ import unicode_literals
+from __future__ import unicode_literals, absolute_import
 
 import os
 from unittest import TestCase
@@ -17,7 +17,7 @@ class TestLockingCopy(TestCase):
         self.tmp = TempIO()
         self.src = self.tmp.mkdir(u'src')
         self.dst = self.tmp.mkdir(u'dst')
-        self.src_file = self.src.putfile(u'somefile', b'')
+        self.src_file = self.src.putfile('somefile', '')
 
     def test_normal_copy_succeeds(self):
         files.locking_copy(self.src_file, self.dst)
