@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2019 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -25,12 +25,14 @@ Utilities
 """
 
 from __future__ import unicode_literals, absolute_import
+from __future__ import division
 
 import sys
 import datetime
 import decimal
 import subprocess
 
+import six
 from pkg_resources import iter_entry_points
 
 try:
@@ -56,6 +58,20 @@ def capture_output(command):
     return output
 
 
+def data_diffs(local_data, host_data, fields=None):
+    """
+    Find all (relevant) fields which differ between the host and local data
+    values for a given record.
+    """
+    if fields is None:
+        fields = list(local_data)
+    diffs = []
+    for field in fields:
+        if local_data[field] != host_data[field]:
+            diffs.append(field)
+    return diffs
+
+
 def import_module_path(module_path):
     """
     Import an arbitrary Python module.
@@ -79,6 +95,14 @@ def import_module_path(module_path):
     return last_module(module, module_path)
 
 
+def get_object_spec(obj):
+    """
+    Returns the "object spec" string for the given object.
+    """
+    # TODO: this only handles a class *instance* currently
+    return '{}:{}'.format(obj.__module__, obj.__class__.__name__)
+
+
 def load_object(specifier):
     """
     Load an arbitrary object from a module, according to a specifier.
@@ -142,15 +166,16 @@ def pretty_boolean(value):
     return "Yes" if value else "No"
 
 
-def hours_as_decimal(hours):
+def hours_as_decimal(hours, places=2):
     """
     Convert the given ``datetime.timedelta`` object into a Decimal whose
     value is in terms of hours.
     """
     if hours is None:
         return
-    minutes = (hours.days * 1440) + (hours.seconds / 60)
-    return decimal.Decimal('{:0.2f}'.format(minutes / 60.0))
+    minutes = (hours.days * 1440) + (hours.seconds // 60)
+    fmt = '{{:0.{}f}}'.format(places)
+    return decimal.Decimal(fmt.format(minutes / 60.0))
 
 
 def pretty_hours(hours=None, seconds=None):
@@ -163,8 +188,23 @@ def pretty_hours(hours=None, seconds=None):
         return ''
     if hours is None:
         hours = datetime.timedelta(seconds=seconds)
+
+    # determine if hours value is positive or negative.  seems like some of the
+    # math can be "off" if negative values are involved, in which case we'll
+    # convert to positive for sake of math and then prefix result with a hyphen
+    negative = False
+
+    if isinstance(hours, decimal.Decimal):
+        if hours < 0:
+            negative = True
+            hours = -hours
+        hours = datetime.timedelta(seconds=int(hours * 3600))
+
     minutes = (hours.days * 1440) + (hours.seconds / 60)
-    return '{}:{:02d}'.format(minutes // 60, minutes % 60)
+    rendered = "{}:{:02d}".format(int(minutes // 60), int(minutes % 60))
+    if negative:
+        rendered = "-{}".format(rendered)
+    return rendered
 
 
 def pretty_quantity(value, empty_zero=False):
@@ -178,8 +218,8 @@ def pretty_quantity(value, empty_zero=False):
         value = int(value)
         if empty_zero and value == 0:
             return ''
-        return str(value)
-    return str(value).rstrip('0')
+        return six.text_type(value)
+    return six.text_type(value).rstrip('0')
 
 
 def progress_loop(func, items, factory, *args, **kwargs):
@@ -210,7 +250,25 @@ def progress_loop(func, items, factory, *args, **kwargs):
             canceled = True
             break
     if prog:
-        prog.destroy()
+        prog.finish()
     if canceled and not allow_cancel:
         raise RuntimeError("Operation was canceled")
     return not canceled
+
+
+def simple_error(error):
+    """
+    Return a "simple" string for the given error.  Result will look like::
+
+       "ErrorClass: Description for the error"
+
+    However the logic checks to ensure the error has a descriptive message
+    first; if it doesn't the result will just be::
+
+       "ErrorClass"
+    """
+    cls = six.text_type(type(error).__name__)
+    msg = six.text_type(error)
+    if msg:
+        return "{}: {}".format(cls, msg)
+    return cls
